"""Robyn route registrations for the user flows."""

from robyn import Request, Response, Robyn
from starlette import status

from ...domain.users import UserFlat
from ...infrastructure.application import json_response
from ...operational import authentication as auth_ops
from ...operational import users as users_ops
from .._helpers import parse_body
from .contracts import (
    ActivationBody,
    EmailChangeConfirmBody,
    EmailChangeRequestBody,
    PasswordChangeBody,
    PasswordResetConfirmBody,
    PasswordResetRequestBody,
    UserCreateBody,
    UserExternalBody,
    UserPublic,
)


def _serialize(user: UserFlat) -> dict:
    return UserPublic.model_validate(user).model_dump(by_alias=True)


def register(app: Robyn) -> None:
    @app.post("/users")
    async def user_create(request: Request) -> Response:
        payload = await parse_body(request, UserCreateBody)
        user = await users_ops.create(payload=payload.model_dump())
        return json_response(
            payload=_serialize(user),
            status_code=status.HTTP_201_CREATED,
        )

    @app.post("/users/external")
    async def user_create_external(request: Request) -> Response:
        payload = await parse_body(request, UserExternalBody)
        user = await users_ops.create_external(payload=payload.model_dump())
        return json_response(
            payload=_serialize(user),
            status_code=status.HTTP_201_CREATED,
        )

    @app.post("/users/activate")
    async def user_activate(request: Request) -> Response:
        payload = await parse_body(request, ActivationBody)
        user = await users_ops.activate(key=payload.key)
        return json_response(
            payload=_serialize(user),
            status_code=status.HTTP_200_OK,
        )

    @app.post("/users/password/change", auth_required=True)
    async def user_password_change(request: Request) -> Response:
        payload = await parse_body(request, PasswordChangeBody)
        user_id = auth_ops.require_user_id(request)
        user = await users_ops.get(user_id=user_id)
        updated = await users_ops.password_change(
            user=user,
            old_password=payload.old_password,
            new_password=payload.new_password,
        )
        return json_response(
            payload=_serialize(updated),
            status_code=status.HTTP_200_OK,
        )

    @app.post("/users/password/reset/request")
    async def user_password_reset_request(request: Request) -> Response:
        payload = await parse_body(request, PasswordResetRequestBody)
        await users_ops.password_reset(email=payload.email)
        return json_response(
            payload={},
            status_code=status.HTTP_202_ACCEPTED,
        )

    @app.post("/users/password/reset/confirm")
    async def user_password_reset_confirm(request: Request) -> Response:
        payload = await parse_body(request, PasswordResetConfirmBody)
        user = await users_ops.password_reset_change(
            key=payload.key,
            new_password=payload.password,
        )
        return json_response(
            payload=_serialize(user),
            status_code=status.HTTP_200_OK,
        )

    @app.post("/users/email-change/request", auth_required=True)
    async def user_email_change_request(request: Request) -> Response:
        payload = await parse_body(request, EmailChangeRequestBody)
        user_id = auth_ops.require_user_id(request)
        user = await users_ops.get(user_id=user_id)
        await users_ops.email_change_request(user=user, email=payload.email)
        return json_response(
            payload={},
            status_code=status.HTTP_202_ACCEPTED,
        )

    @app.post("/users/email-change/confirm")
    async def user_email_change_confirm(request: Request) -> Response:
        payload = await parse_body(request, EmailChangeConfirmBody)
        user = await users_ops.email_change_confirmation(key=payload.key)
        return json_response(
            payload=_serialize(user),
            status_code=status.HTTP_200_OK,
        )



import uuid
from typing import Callable

from pydantic import EmailStr

from ...config import settings
from ...infrastructure.application import DatabaseError, UnprocessableError
from ...infrastructure.authentication import AuthProvider
from ...infrastructure.database import transaction
from .constants import Role
from .entities import UserFlat, UserUncommitted
from .repository import UsersRepository


def generate_key(email: EmailStr) -> uuid.UUID:
    return uuid.uuid3(namespace=uuid.uuid4(), name=email)


def create_activation_link(activation_key: uuid.UUID) -> str:
    return f"{settings.integrations.frontend.activation_base_url}/{activation_key}"


def create_password_reset_link(activation_key: uuid.UUID) -> str:
    return f"{settings.integrations.frontend.password_reset_base_url}/{activation_key}"


def create_email_change_link(activation_key: uuid.UUID) -> str:
    return f"{settings.integrations.frontend.email_change_base_url}/{activation_key}"


async def create(
    payload: dict, repository_factory: Callable[[], UsersRepository]
) -> UserFlat:
    payload.setdefault("role", Role.ADMIN)
    try:
        async with transaction():
            repository = repository_factory()
            schema = UserUncommitted(**payload)
            user = await repository.create(schema)
    except DatabaseError as exc:
        raise UnprocessableError(message="User already exists") from exc
    return user


async def update_partial(
    id_: int,
    payload: dict,
    repository_factory: Callable[[], UsersRepository],
) -> UserFlat:
    async with transaction():
        repository = repository_factory()
        return await repository.update(attr="id", value=id_, payload=payload)


async def activate(
    id_: int, repository_factory: Callable[[], UsersRepository]
) -> UserFlat:
    async with transaction():
        repository = repository_factory()
        return await repository.update(
            attr="id", value=id_, payload={"is_active": True}
        )


async def password_update(
    id_: int,
    password_hash: str,
    repository_factory: Callable[[], UsersRepository],
) -> UserFlat:
    payload = {
        "password": password_hash,
        "auth_provider": AuthProvider.INTERNAL,
    }
    async with transaction():
        repository = repository_factory()
        return await repository.update(attr="id", value=id_, payload=payload)



import uuid
from datetime import datetime, timedelta

from pydantic import EmailStr, Field, field_validator, model_validator

from ..schemas import PublicEntity
from ..utils import UnprocessableError

# Authentication Contracts


class LoginRequestBody(PublicEntity):
    login: EmailStr | str = Field(
        description="User login", examples=["john@email.com", "john"]
    )
    password: str = Field(description="User password", examples=["password"])


class TokenResponse(PublicEntity):
    access_token: str = Field(description="Access token", examples=["token"])
    token_type: str = Field(
        description="Token type", examples=["Bearer"], default="Bearer"
    )


class TokenInfo(PublicEntity):
    subject: str = Field(description="Subject", examples=["1"])
    email: EmailStr = Field(
        description="User email", examples=["john@email.com"]
    )
    username: str = Field(description="User username", examples=["john"])
    issued_at: datetime = Field(
        description="Issued at", examples=[datetime.now()]
    )
    expires_at: datetime = Field(
        description="Expires at",
        examples=[datetime.now() + timedelta(hours=1)],
    )


# User Contracts


class _BaseUser(PublicEntity):
    username: str = Field(
        description="User username",
        examples=["john"],
        min_length=1,
        max_length=255,
    )
    email: EmailStr = Field(
        description="User email",
        examples=["john@email.com"],
    )


class UserCreateBody(_BaseUser):
    password: str = Field(
        description="User password",
        examples=["@Dm1n#LKJ"],
    )

    @field_validator("password", mode="before")
    @classmethod
    def password_nist(cls, value: str) -> str:
        # Simplified validation for MVC example
        if len(value) < 8:
            raise ValueError("Password must be at least 8 characters long")
        return value

    @model_validator(mode="before")
    @classmethod
    def username_email(cls, values: dict) -> dict:
        if values.get("username") == values.get("email"):
            raise UnprocessableError(
                message="The username and email must be different."
            )
        return values


class UserExternalBody(_BaseUser):
    role: int = Field(
        description="User role",
        examples=[1],
        default=1,
    )
    auth_provider: str | None = Field(
        default=None,
        description="External auth provider identifier",
    )


class ActivationBody(PublicEntity):
    key: uuid.UUID = Field(description="Activation key from email")


class PasswordChangeBody(PublicEntity):
    old_password: str = Field(description="User's current password")
    new_password: str = Field(description="A new user's password")

    @field_validator("new_password", mode="before")
    @classmethod
    def password_nist(cls, value: str) -> str:
        if len(value) < 8:
            raise ValueError("Password must be at least 8 characters long")
        return value


class PasswordResetRequestBody(PublicEntity):
    email: EmailStr = Field(
        description="User email", examples=["john@email.com"]
    )


class PasswordResetConfirmBody(PublicEntity):
    key: uuid.UUID = Field(
        description="Password reset key that is taken from the email"
    )
    password: str = Field(description="A new user's password")

    @field_validator("password", mode="before")
    @classmethod
    def password_nist(cls, value: str) -> str:
        if len(value) < 8:
            raise ValueError("Password must be at least 8 characters long")
        return value


class EmailChangeRequestBody(PublicEntity):
    email: EmailStr = Field(description="A new email you want to change to")


class EmailChangeConfirmBody(PublicEntity):
    key: uuid.UUID = Field(
        description="Email change key that is taken from the email"
    )


class UserPublic(_BaseUser):
    id: int = Field(description="User id", examples=[1])
    is_active: bool = Field(
        description="Whether the user activated the account",
        examples=[True],
    )
    role: int = Field(
        description="User role. Possible values: [1,2,3]", examples=[2]
    )



from typing import Any, AsyncGenerator

from ....domain.users import UsersRepository as UsersRepositoryInterface
from ....domain.users.entities import UserFlat, UserUncommitted
from ..tables import UsersTable
from .base import BaseRepository


class UsersRepository(BaseRepository[UsersTable], UsersRepositoryInterface):
    async def all(self) -> AsyncGenerator[UserFlat, None]:
        async for instance in self._all():
            yield UserFlat.model_validate(instance)

    async def get(self, id_: int) -> UserFlat:
        instance = await self._get(key="id", value=id_)
        return UserFlat.model_validate(instance)

    async def get_by_login(self, login: str) -> UserFlat:
        instance = await self._get(key="username", value=login)
        return UserFlat.model_validate(instance)

    async def create(self, schema: UserUncommitted) -> UserFlat:
        instance = await self._save(schema.model_dump())
        return UserFlat.model_validate(instance)

    async def update(
        self, attr: str, value: Any, payload: dict[str, Any]
    ) -> UserFlat:
        schema = await self._update(attr, value, payload)
        return UserFlat.model_validate(schema)



from contextlib import asynccontextmanager
from typing import AsyncGenerator

from loguru import logger
from sqlalchemy.exc import IntegrityError, InvalidRequestError
from sqlalchemy.ext.asyncio import AsyncSession

from ....infrastructure.application import DatabaseError
from .session import CTX_SESSION, create_session


@asynccontextmanager
async def transaction() -> AsyncGenerator[AsyncSession, None]:
    session = create_session()
    CTX_SESSION.set(session)
    try:
        yield session
        await session.commit()
    except DatabaseError as exc:
        logger.error(f"Rolling back changes: {exc}")
        await session.rollback()
        raise
    except (IntegrityError, InvalidRequestError) as exc:  # pragma: no cover
        logger.error(f"Rolling back changes: {exc}")
        await session.rollback()
        raise DatabaseError(message=str(exc)) from exc
    finally:
        await session.close()



import logging
from pathlib import Path

from app.config import settings
from app.middlewares import cors, sessions
from app.urls import register_routes
from app.utils import error_response
from app.views.authentication import JWTAuthenticationHandler
from loguru import logger
from robyn import Robyn

log_path = Path(settings.root_dir) / "logs"
log_path.mkdir(exist_ok=True)
logger.add(
    log_path / f"{settings.logging.file}.log",
    format=settings.logging.format,
    rotation=settings.logging.rotation,
    compression=settings.logging.compression,
    level="INFO",
)

if not settings.debug:
    logging.getLogger("sqlalchemy.engine").setLevel(logging.WARNING)
    logging.getLogger("alembic").setLevel(logging.WARNING)

app = Robyn(__file__)

# Configure app
app.exception(error_response)
app.configure_authentication(JWTAuthenticationHandler())

# Middlewares
sessions.register(app)
cors.register(app)

# Register routes
register_routes(app)

if __name__ == "__main__":
    app.start(host="0.0.0.0", port=8000)



"""Internal error hierarchy, adapted from src 2."""

from starlette import status

__all__ = (
    "BaseError",
    "BadRequestError",
    "UnprocessableError",
    "NotFoundError",
    "AuthenticationError",
    "AuthorizationError",
    "DatabaseError",
)


class BaseError(Exception):
    def __init__(
        self,
        *,
        message: str = "",
        status_code: int = status.HTTP_500_INTERNAL_SERVER_ERROR,
    ) -> None:
        self.message = message or self.__class__.__name__
        self.status_code = status_code
        super().__init__(self.message)


class BadRequestError(BaseError):
    def __init__(self, *, message: str = "Bad request") -> None:
        super().__init__(
            message=message, status_code=status.HTTP_400_BAD_REQUEST
        )


class UnprocessableError(BaseError):
    def __init__(self, *, message: str = "Validation error") -> None:
        super().__init__(
            message=message, status_code=status.HTTP_422_UNPROCESSABLE_ENTITY
        )


class NotFoundError(BaseError):
    def __init__(self, *, message: str = "Not found") -> None:
        super().__init__(
            message=message, status_code=status.HTTP_404_NOT_FOUND
        )


class AuthenticationError(BaseError):
    def __init__(self, *, message: str = "Not authenticated") -> None:
        super().__init__(
            message=message, status_code=status.HTTP_401_UNAUTHORIZED
        )


class AuthorizationError(BaseError):
    def __init__(self, *, message: str = "Forbidden") -> None:
        super().__init__(
            message=message, status_code=status.HTTP_403_FORBIDDEN
        )


class DatabaseError(BaseError):
    def __init__(self, *, message: str = "Database error") -> None:
        super().__init__(
            message=message, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
        )



from pydantic import AnyHttpUrl, BaseModel


class Settings(BaseModel):
    login_url: AnyHttpUrl = AnyHttpUrl("http://frontend.local/login")
    activation_base_url: AnyHttpUrl = AnyHttpUrl(
        "http://frontend.local/activate"
    )
    password_reset_base_url: AnyHttpUrl = AnyHttpUrl(
        "http://frontend.local/password-reset"
    )
    email_change_base_url: AnyHttpUrl = AnyHttpUrl(
        "http://frontend.local/email-change"
    )



from datetime import datetime
from typing import Generic, TypeVar

from pydantic import Field

from ..application import InternalEntity

_CacheEntryInstance = TypeVar("_CacheEntryInstance", bound=InternalEntity)


class CacheEntry(InternalEntity, Generic[_CacheEntryInstance]):
    instance: _CacheEntryInstance
    created_at: datetime = Field(default_factory=datetime.utcnow)



import asyncio
import functools
from contextlib import asynccontextmanager
from contextvars import ContextVar
from typing import Any, AsyncGenerator, Sequence

from loguru import logger
from tortoise import Tortoise, connections
from tortoise.backends.base.client import BaseDBAsyncClient
from tortoise.exceptions import IntegrityError, OperationalError
from tortoise.transactions import in_transaction

from ..config import settings
from ..utils import BaseError


class DatabaseError(BaseError):
    def __init__(self, message: str = "Database error"):
        super().__init__(message)


APP_LABEL = "models"
MODEL_MODULES = (
    "app.models.models",
    "aerich.models",
)
DRIVER_ALIASES = {
    "sqlite+aiosqlite": "sqlite",
    "postgresql+asyncpg": "postgres",
    "mysql+aiomysql": "mysql",
}

_ENGINE_LOCK = asyncio.Lock()
_INITIALIZED = False


def _normalize_database_url(url: str) -> str:
    for candidate, alias in DRIVER_ALIASES.items():
        if url.startswith(candidate):
            return url.replace(candidate, alias, 1)
    return url


def _build_config() -> dict[str, Any]:
    return {
        "connections": {
            "default": _normalize_database_url(settings.database.url)
        },
        "apps": {
            APP_LABEL: {
                "models": list(MODEL_MODULES),
                "default_connection": "default",
            }
        },
        "use_tz": False,
        "timezone": "UTC",
    }


@functools.lru_cache(maxsize=1)
def build_engine() -> dict[str, Any]:
    return _build_config()


TORTOISE_ORM = build_engine()


async def create_engine() -> dict[str, Any]:
    global _INITIALIZED
    if _INITIALIZED:
        return TORTOISE_ORM

    async with _ENGINE_LOCK:
        if not _INITIALIZED:
            await Tortoise.init(config=TORTOISE_ORM)
            _INITIALIZED = True
    return TORTOISE_ORM


async def close_engine() -> None:
    global _INITIALIZED
    if not _INITIALIZED:
        return
    await Tortoise.close_connections()
    _INITIALIZED = False


@asynccontextmanager
async def transaction() -> AsyncGenerator[BaseDBAsyncClient, None]:
    await create_engine()
    async with in_transaction(connection_name="default") as connection:
        try:
            yield connection
        except DatabaseError as exc:
            logger.error(f"Rolling back changes: {exc}")
            raise
        except (IntegrityError, OperationalError) as exc:
            logger.error(f"Rolling back changes: {exc}")
            raise DatabaseError(message=str(exc)) from exc


CTX_CONNECTION: ContextVar[BaseDBAsyncClient | None] = ContextVar(
    "tortoise_connection", default=None
)


class Session:
    _ERRORS = (OperationalError,)

    def __init__(self) -> None:
        connection = CTX_CONNECTION.get()
        if connection is None:
            # Fallback to default connection if not in transaction context
            # This is a simplification for MVC; ideally we should be in a transaction
            try:
                connection = connections.get("default")
            except KeyError:
                pass

        if connection is None:
            # If still None, we might not be initialized or in a weird state
            # For now, let's try to get it again or raise
            pass

        self._connection: BaseDBAsyncClient = connection
        # Maintain SQLAlchemy parity for downstream consumers.
        self._session: BaseDBAsyncClient = connection

    async def execute(
        self, query: str, values: Sequence[Any] | None = None
    ) -> list[dict[str, Any]]:
        try:
            return await self._connection.execute_query_dict(query, values)
        except self._ERRORS as exc:
            raise DatabaseError(message=str(exc)) from exc

    @property
    def connection(self) -> BaseDBAsyncClient:
        return self._connection



from typing import Any

import msgspec
from robyn import Response
from starlette import status

JSON_HEADERS = {"content-type": "application/json; charset=utf-8"}


class BaseError(Exception):
    def __init__(
        self, message: str = "", status_code: int = status.HTTP_400_BAD_REQUEST
    ):
        self.message = message
        self.status_code = status_code
        super().__init__(message)


class BadRequestError(BaseError):
    def __init__(self, message: str = "Bad request"):
        super().__init__(message, status_code=status.HTTP_400_BAD_REQUEST)


class AuthenticationError(BaseError):
    def __init__(self, message: str = "Authentication failed"):
        super().__init__(message, status_code=status.HTTP_401_UNAUTHORIZED)


class NotFoundError(BaseError):
    def __init__(self, message: str = "Not found"):
        super().__init__(message, status_code=status.HTTP_404_NOT_FOUND)


class UnprocessableError(BaseError):
    def __init__(self, message: str = "Unprocessable entity"):
        super().__init__(
            message, status_code=status.HTTP_422_UNPROCESSABLE_ENTITY
        )


class DatabaseError(BaseError):
    def __init__(self, message: str = "Database error"):
        super().__init__(
            message, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
        )


def _encode(payload: Any) -> bytes:
    return msgspec.json.encode(payload)


def json_response(
    payload: Any, status_code: int = status.HTTP_200_OK
) -> Response:
    return Response(
        status_code,
        JSON_HEADERS,
        _encode({"result": payload}),
    )


def error_response(exc: Exception) -> Response:
    if isinstance(exc, BaseError):
        payload = {"message": exc.message}
        body = _encode({"error": payload})
        return Response(exc.status_code, JSON_HEADERS, body)

    payload = {"message": str(exc) or "Internal Server Error"}
    body = _encode({"error": payload})
    return Response(status.HTTP_500_INTERNAL_SERVER_ERROR, JSON_HEADERS, body)



from typing import Literal

from pydantic import AnyHttpUrl, BaseModel


class Settings(BaseModel):
    allow_origins: list[AnyHttpUrl | Literal["*"]] = ["*"]
    allow_methods: list[str] = ["*"]
    allow_headers: list[str] = ["*"]
    allow_credentials: bool = True
    expose_headers: list[str] = []
    max_age: int = 600



from enum import StrEnum, auto

from passlib.context import CryptContext


class AuthProvider(StrEnum):
    INTERNAL = auto()
    GOOGLE = auto()
    MICROSOFT = auto()
    META = auto()


pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")



from __future__ import annotations

import asyncio
from logging.config import fileConfig
from typing import Mapping

from alembic import context
from app.config import settings
from app.models.models import Base
from sqlalchemy import pool
from sqlalchemy.ext.asyncio import async_engine_from_config

config = context.config

if config.config_file_name is not None:
    fileConfig(config.config_file_name)

target_metadata = Base.metadata


def _configure_section() -> Mapping[str, str]:
    section = config.get_section(config.config_ini_section) or {}
    section["sqlalchemy.url"] = settings.database.url
    return section


def run_migrations_offline() -> None:
    context.configure(
        url=settings.database.url,
        target_metadata=target_metadata,
        literal_binds=True,
        compare_type=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    connectable = async_engine_from_config(
        _configure_section(),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    async def _run_async_migrations() -> None:
        async with connectable.connect() as connection:
            await connection.run_sync(_do_run_migrations)
        await connectable.dispose()

    def _do_run_migrations(connection):
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            compare_type=True,
        )
        with context.begin_transaction():
            context.run_migrations()

    asyncio.run(_run_async_migrations())


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



from __future__ import annotations

from contextvars import ContextVar
from typing import Any, Sequence

from tortoise import connections
from tortoise.backends.base.client import BaseDBAsyncClient
from tortoise.exceptions import OperationalError

from ....infrastructure.application import DatabaseError
from .engine import create_engine

CTX_CONNECTION: ContextVar[BaseDBAsyncClient | None] = ContextVar(
    "tortoise_connection", default=None
)


async def create_session(
    connection: BaseDBAsyncClient | None = None,
) -> BaseDBAsyncClient:
    await create_engine()
    conn = connection
    if conn is None:
        try:
            conn = connections.get("default")
        except KeyError as exc:  # pragma: no cover - defensive
            raise DatabaseError(
                message="Unable to acquire Tortoise connection"
            ) from exc
    CTX_CONNECTION.set(conn)
    return conn


class Session:
    _ERRORS = (OperationalError,)

    def __init__(self) -> None:
        connection = CTX_CONNECTION.get()
        if connection is None:
            raise DatabaseError(
                message="Session requested outside transaction context"
            )
        self._connection: BaseDBAsyncClient = connection
        # Maintain SQLAlchemy parity for downstream consumers.
        self._session: BaseDBAsyncClient = connection

    async def execute(
        self, query: str, values: Sequence[Any] | None = None
    ) -> list[dict[str, Any]]:
        try:
            return await self._connection.execute_query_dict(query, values)
        except self._ERRORS as exc:  # pragma: no cover - defensive
            raise DatabaseError(message=str(exc)) from exc

    @property
    def connection(self) -> BaseDBAsyncClient:
        return self._connection



from contextvars import ContextVar

from sqlalchemy.exc import IntegrityError, InvalidRequestError
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession

from ....infrastructure.application import DatabaseError
from .engine import create_engine


def create_session(engine: AsyncEngine | None = None) -> AsyncSession:
    return AsyncSession(
        engine or create_engine(), expire_on_commit=False, autoflush=False
    )


CTX_SESSION: ContextVar[AsyncSession] = ContextVar(
    "session", default=create_session()
)


class Session:
    _ERRORS = (IntegrityError, InvalidRequestError)

    def __init__(self) -> None:
        self._session: AsyncSession = CTX_SESSION.get()

    async def execute(self, query):
        try:
            return await self._session.execute(query)
        except self._ERRORS as exc:  # pragma: no cover - defensive
            raise DatabaseError(message=str(exc)) from exc



import functools

from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine

from ....config import settings


def build_engine() -> AsyncEngine:
    return create_async_engine(
        settings.database.url,
        future=True,
        pool_pre_ping=True,
        echo=settings.debug,
    )


@functools.lru_cache(maxsize=1)
def create_engine() -> AsyncEngine:
    return build_engine()



"""Helpers for Robyn handlers."""

from typing import Type, TypeVar

import msgspec
from pydantic import BaseModel, ValidationError
from robyn import Request

from ..infrastructure.application import BadRequestError

T = TypeVar("T", bound=BaseModel)


def _clean_body(body: bytes | str | None) -> bytes:
    if body is None:
        return b"{}"
    raw = body.encode() if isinstance(body, str) else body
    return raw or b"{}"


async def parse_body(request: Request, model: Type[T]) -> T:
    try:
        payload = msgspec.json.decode(_clean_body(request.body))
    except msgspec.DecodeError as exc:
        raise BadRequestError(message=f"JSON is malformed: {exc}") from exc

    if not isinstance(payload, dict):
        raise BadRequestError(message="Request body must be an object")

    try:
        return model(**payload)
    except ValidationError as exc:
        raise BadRequestError(message=exc.errors()[0]["msg"]) from exc


def _normalized_headers(request: Request) -> dict[str, str]:
    raw_headers = getattr(request, "headers", None) or {}
    if isinstance(raw_headers, dict):
        items = raw_headers.items()
    else:
        try:
            items = dict(raw_headers).items()  # type: ignore[arg-type]
        except Exception:  # pragma: no cover - defensive
            try:
                items = raw_headers.items()  # type: ignore[assignment,call-arg]
            except Exception:
                return {}

    headers: dict[str, str] = {}
    for key, value in items:
        if isinstance(key, bytes):
            key = key.decode()
        if isinstance(value, bytes):
            value = value.decode()
        headers[str(key).lower()] = str(value)
    return headers


def get_header(request: Request, name: str) -> str | None:
    return _normalized_headers(request).get(name.lower())



from email.message import EmailMessage as SMTPEmailMessage
from email.utils import formataddr

import aiosmtplib
from loguru import logger

from ...config import settings
from .entities import EmailMessage


class MailingService:
    def __init__(self) -> None:
        self._config = settings.mailing

    def _build_message(self, message: EmailMessage) -> SMTPEmailMessage:
        email = SMTPEmailMessage()
        sender = (
            formataddr((self._config.sender_name, self._config.sender_email))
            if self._config.sender_name
            else str(self._config.sender_email)
        )
        email["From"] = sender
        email["To"] = ", ".join(message.recipients)
        email["Subject"] = message.subject
        email.set_content(message.body)
        return email

    async def send(
        self, message: EmailMessage
    ) -> None:  # pragma: no cover - relies on I/O
        smtp_message = self._build_message(message)
        logger.info(
            "Sending email via SMTP -> host=%s port=%s recipients=%s subject=%s",
            self._config.host,
            self._config.port,
            ",".join(message.recipients),
            message.subject,
        )
        await aiosmtplib.send(
            smtp_message,
            hostname=self._config.host,
            port=self._config.port,
            start_tls=self._config.start_tls,
            username=self._config.username,
            password=self._config.password,
        )


mailing_service = MailingService()



from pydantic import BaseModel, EmailStr


class Settings(BaseModel):
    host: str = "localhost"
    port: int = 1025
    username: str | None = None
    password: str | None = None
    start_tls: bool = False
    sender_email: EmailStr = "no-reply@example.com"
    sender_name: str | None = "Robyn App"



from typing import Any, AsyncGenerator, Generic, get_args, get_origin

from sqlalchemy import asc, delete, desc, func, select, update
from sqlalchemy.engine import Result

from ....infrastructure.application import (
    DatabaseError,
    NotFoundError,
    UnprocessableError,
)
from ..services.session import Session
from ..tables import ConcreteTable


class BaseRepository(Session, Generic[ConcreteTable]):
    schema_class: type[ConcreteTable] | None = None

    def __init__(self) -> None:
        super().__init__()
        if not getattr(self, "schema_class", None):
            raise UnprocessableError(message="schema_class is required")

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if cls is BaseRepository:
            return
        if cls.schema_class is None:
            for base in getattr(cls, "__orig_bases__", ()):
                if get_origin(base) is BaseRepository:
                    args = get_args(base)
                    if args:
                        cls.schema_class = args[0]
                        break
        if cls.schema_class is None:
            raise UnprocessableError(message="schema_class is required")

    async def _filter(self, **filters: Any) -> ConcreteTable:
        query = select(self.schema_class).filter_by(**filters)
        result: Result = await self.execute(query)
        schema = result.scalar_one_or_none()
        if not schema:
            raise NotFoundError
        return schema

    async def _update(
        self, key: str, value: Any, payload: dict[str, Any]
    ) -> ConcreteTable:
        query = (
            update(self.schema_class)
            .where(getattr(self.schema_class, key) == value)
            .values(payload)
            .returning(self.schema_class)
        )
        result: Result = await self.execute(query)
        await self._session.flush()

        schema = result.scalar_one_or_none()
        if not schema:
            raise DatabaseError
        return schema

    async def _get(self, key: str, value: Any) -> ConcreteTable:
        query = select(self.schema_class).where(
            getattr(self.schema_class, key) == value
        )
        result: Result = await self.execute(query)
        schema = result.scalars().one_or_none()
        if not schema:
            raise NotFoundError
        return schema

    async def count(self) -> int:
        result: Result = await self.execute(func.count(self.schema_class.id))
        value = result.scalar()
        if not isinstance(value, int):  # pragma: no cover - sanity
            raise UnprocessableError(message=f"Count returned {value}")
        return value

    async def _first(self, by: str = "id") -> ConcreteTable:
        result: Result = await self.execute(
            select(self.schema_class).order_by(asc(by)).limit(1)
        )
        schema = result.scalar_one_or_none()
        if not schema:
            raise NotFoundError
        return schema

    async def _last(self, by: str = "id") -> ConcreteTable:
        result: Result = await self.execute(
            select(self.schema_class).order_by(desc(by)).limit(1)
        )
        schema = result.scalar_one_or_none()
        if not schema:
            raise NotFoundError
        return schema

    async def _save(self, payload: dict[str, Any]) -> ConcreteTable:
        schema = self.schema_class(**payload)
        self._session.add(schema)
        await self._session.flush()
        await self._session.refresh(schema)
        return schema

    async def _all(self) -> AsyncGenerator[ConcreteTable, None]:
        result: Result = await self.execute(select(self.schema_class))
        for schema in result.scalars().all():
            yield schema

    async def delete(self, id_: int) -> None:
        await self.execute(
            delete(self.schema_class).where(self.schema_class.id == id_)
        )
        await self._session.flush()



import functools
from contextlib import asynccontextmanager
from contextvars import ContextVar
from typing import AsyncGenerator

from loguru import logger
from sqlalchemy.exc import IntegrityError, InvalidRequestError
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)

from ..config import settings
from ..utils import BaseError


class DatabaseError(BaseError):
    def __init__(self, message: str = "Database error"):
        super().__init__(message)


def build_engine() -> AsyncEngine:
    return create_async_engine(
        settings.database.url,
        future=True,
        pool_pre_ping=True,
        echo=settings.debug,
    )


@functools.lru_cache(maxsize=1)
def create_engine() -> AsyncEngine:
    return build_engine()


def create_session() -> AsyncSession:
    Session = async_sessionmaker(
        create_engine(),
        expire_on_commit=False,
        class_=AsyncSession,
    )
    return Session()


@asynccontextmanager
async def transaction() -> AsyncGenerator[AsyncSession, None]:
    session = create_session()
    try:
        yield session
        await session.commit()
    except DatabaseError as exc:
        logger.error(f"Rolling back changes: {exc}")
        await session.rollback()
        raise
    except (IntegrityError, InvalidRequestError) as exc:
        logger.error(f"Rolling back changes: {exc}")
        await session.rollback()
        raise DatabaseError(message=str(exc)) from exc
    finally:
        await session.close()


CTX_SESSION: ContextVar[AsyncSession] = ContextVar(
    "session", default=create_session()
)


class Session:
    _ERRORS = (IntegrityError, InvalidRequestError)

    def __init__(self) -> None:
        self._session: AsyncSession = CTX_SESSION.get()

    async def execute(self, query):
        try:
            return await self._session.execute(query)
        except self._ERRORS as exc:
            raise DatabaseError(message=str(exc)) from exc



from datetime import datetime, timedelta, timezone

import jwt
from robyn import Request, Response, Robyn
from robyn.authentication import AuthenticationHandler, BearerGetter, Identity
from starlette import status

from ..authentication import pwd_context
from ..config import settings
from ..models import UsersRepository, transaction
from ..utils import AuthenticationError, NotFoundError, json_response
from .contracts import LoginRequestBody, TokenInfo, TokenResponse
from .helpers import parse_body


def create_access_token(user_id: int, email: str, username: str) -> str:
    now = datetime.now(timezone.utc)
    ttl = settings.authentication.access_token.ttl
    exp = now + timedelta(seconds=ttl)
    payload = {
        "sub": str(user_id),
        "email": email,
        "username": username,
        "iat": int(now.timestamp()),
        "exp": int(exp.timestamp()),
    }
    return jwt.encode(
        payload,
        settings.authentication.access_token.secret_key,
        algorithm=settings.authentication.algorithm,
    )


def decode_access_token(token: str) -> dict:
    try:
        payload = jwt.decode(
            token,
            settings.authentication.access_token.secret_key,
            algorithms=[settings.authentication.algorithm],
        )
    except jwt.ExpiredSignatureError as exc:
        raise AuthenticationError(message="Token has expired") from exc
    except jwt.PyJWTError as exc:
        raise AuthenticationError(message="Invalid token") from exc
    return payload


def require_user_id(request: Request) -> int:
    identity: Identity | None = getattr(request, "identity", None)
    if identity is None:
        raise AuthenticationError(message="Authentication required")

    sub = identity.claims.get("sub")
    if sub is None:
        raise AuthenticationError(message="Subject claim missing")

    try:
        return int(sub)
    except (TypeError, ValueError) as exc:
        raise AuthenticationError(message="Subject claim invalid") from exc


class JWTAuthenticationHandler(AuthenticationHandler):
    def __init__(self) -> None:
        super().__init__(BearerGetter())
        self._last_error = AuthenticationError()

    def authenticate(self, request) -> Identity | None:
        token = self.token_getter.get_token(request)
        if not token:
            self._last_error = AuthenticationError(
                message="Token not provided"
            )
            return None

        try:
            claims = decode_access_token(token)
        except AuthenticationError as exc:
            self._last_error = exc
            return None

        normalized = {str(key): str(value) for key, value in claims.items()}
        return Identity(claims=normalized)

    @property
    def unauthorized_response(self):
        # Simplified error response
        return json_response(
            {"message": self._last_error.message},
            status_code=self._last_error.status_code,
        )


def register(app: Robyn) -> None:
    @app.post("/auth/login")
    async def login(request: Request) -> Response:
        payload = await parse_body(request, LoginRequestBody)

        async with transaction():
            repo = UsersRepository()
            try:
                user = await repo.get_by_login(login=payload.login)
            except NotFoundError:
                raise AuthenticationError(message="Invalid credentials")

        if not user.is_active:
            raise AuthenticationError(message="Account not activated")

        if not user.password:
            raise AuthenticationError(message="Password not set")
        if not pwd_context.verify(
            payload.password, user.password, scheme="bcrypt"
        ):
            raise AuthenticationError(message="Invalid credentials")

        token = create_access_token(user.id, user.email, user.username)
        response = TokenResponse(access_token=token)
        return json_response(
            payload=response.model_dump(by_alias=True),
            status_code=status.HTTP_200_OK,
        )

    @app.get("/auth/me", auth_required=True)
    async def me(request: Request) -> Response:
        identity: Identity | None = getattr(request, "identity", None)
        if identity is None:
            raise AuthenticationError(message="Authentication required")

        claims = identity.claims
        issued_at = datetime.fromtimestamp(
            float(claims.get("iat", "0")), tz=timezone.utc
        )
        expires_at = datetime.fromtimestamp(
            float(claims.get("exp", "0")), tz=timezone.utc
        )
        info = TokenInfo(
            subject=claims.get("sub", ""),
            email=claims.get("email", ""),
            username=claims.get("username", ""),
            issued_at=issued_at,
            expires_at=expires_at,
        )
        return json_response(
            payload=info.model_dump(by_alias=True),
            status_code=status.HTTP_200_OK,
        )



from pydantic import BaseModel

from .core import ROOT_PATH


class Settings(BaseModel):
    driver: str = "sqlite+aiosqlite"
    host: str = "database"
    port: int = 5432
    user: str = "sqlite"
    password: str = "sqlite"
    name: str = "database"

    @property
    def url(self) -> str:
        # SQLite stays the default for dev/test simplicity.
        if "sqlite" in self.driver:
            return f"{self.driver}:///{ROOT_PATH / self.name}.db"
        return (
            f"{self.driver}://{self.user}:{self.password}"
            f"@{self.host}:{self.port}/{self.name}"
        )



from typing import TypeVar

from tortoise import fields
from tortoise.models import Model

from ..authentication import AuthProvider

__all__ = (
    "BaseTable",
    "ConcreteTable",
    "UsersTable",
    "NewsLetterSubscriptionsTable",
)


class BaseTable(Model):
    id = fields.IntField(pk=True)
    created_at = fields.DatetimeField(auto_now_add=True)
    updated_at = fields.DatetimeField(auto_now=True)

    class Meta:
        abstract = True


ConcreteTable = TypeVar("ConcreteTable", bound="BaseTable")


class UsersTable(BaseTable):
    username = fields.CharField(max_length=255, unique=True, index=True)
    email = fields.CharField(max_length=255, unique=True, index=True)
    password = fields.CharField(max_length=512, null=True)
    is_active = fields.BooleanField(default=False)
    role = fields.SmallIntField(default=1)
    auth_provider = fields.CharEnumField(
        enum_type=AuthProvider, default=AuthProvider.INTERNAL
    )

    class Meta:
        table = "users"
        ordering = ("id",)


class NewsLetterSubscriptionsTable(BaseTable):
    email = fields.CharField(max_length=255, unique=True, index=True)

    class Meta:
        table = "news_letter_subscriptions"
        ordering = ("id",)



"""Add module for injecting business logic templates into existing projects."""

from .utils import add_business_logic, read_project_config, validate_project

__all__ = [
    "add_business_logic",
    "read_project_config",
    "validate_project",
]



from pydantic import EmailStr, field_validator

from ...infrastructure.application import InternalEntity, TimeStampMixin
from ...infrastructure.authentication import AuthProvider
from .constants import Role


class UserUncommitted(InternalEntity):
    username: str
    email: EmailStr
    password: str
    role: int
    is_active: bool = False
    auth_provider: AuthProvider = AuthProvider.INTERNAL

    @field_validator("role", mode="before")
    @classmethod
    def role_validator(cls, value: int) -> int:
        if value not in Role.values():
            raise ValueError(f"Unsupported role: {value}")
        return value


class UserFlat(UserUncommitted, TimeStampMixin):
    id: int


class PasswordForgot(InternalEntity):
    email: EmailStr


class EmailChange(InternalEntity):
    user_id: int
    email: EmailStr



from typing import Any, AsyncGenerator, Generic, get_args, get_origin

from sqlalchemy import asc, delete, desc, func, select, update
from sqlalchemy.engine import Result

from ..schemas import UserFlat, UserUncommitted
from ..utils import DatabaseError, NotFoundError, UnprocessableError
from .database import Session
from .models import ConcreteTable, UsersTable


class BaseRepository(Session, Generic[ConcreteTable]):
    schema_class: type[ConcreteTable] | None = None

    def __init__(self) -> None:
        super().__init__()
        if not getattr(self, "schema_class", None):
            raise UnprocessableError(message="schema_class is required")

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if cls is BaseRepository:
            return
        if cls.schema_class is None:
            for base in getattr(cls, "__orig_bases__", ()):
                if get_origin(base) is BaseRepository:
                    args = get_args(base)
                    if args:
                        cls.schema_class = args[0]
                        break
        if cls.schema_class is None:
            raise UnprocessableError(message="schema_class is required")

    async def _filter(self, **filters: Any) -> ConcreteTable:
        query = select(self.schema_class).filter_by(**filters)
        result: Result = await self.execute(query)
        schema = result.scalar_one_or_none()
        if not schema:
            raise NotFoundError
        return schema

    async def _update(
        self, key: str, value: Any, payload: dict[str, Any]
    ) -> ConcreteTable:
        query = (
            update(self.schema_class)
            .where(getattr(self.schema_class, key) == value)
            .values(payload)
            .returning(self.schema_class)
        )
        result: Result = await self.execute(query)
        await self._session.flush()

        schema = result.scalar_one_or_none()
        if not schema:
            raise DatabaseError
        return schema

    async def _get(self, key: str, value: Any) -> ConcreteTable:
        query = select(self.schema_class).where(
            getattr(self.schema_class, key) == value
        )
        result: Result = await self.execute(query)
        schema = result.scalars().one_or_none()
        if not schema:
            raise NotFoundError
        return schema

    async def count(self) -> int:
        result: Result = await self.execute(func.count(self.schema_class.id))
        value = result.scalar()
        if not isinstance(value, int):
            raise UnprocessableError(message=f"Count returned {value}")
        return value

    async def _first(self, by: str = "id") -> ConcreteTable:
        result: Result = await self.execute(
            select(self.schema_class).order_by(asc(by)).limit(1)
        )
        schema = result.scalar_one_or_none()
        if not schema:
            raise NotFoundError
        return schema

    async def _last(self, by: str = "id") -> ConcreteTable:
        result: Result = await self.execute(
            select(self.schema_class).order_by(desc(by)).limit(1)
        )
        schema = result.scalar_one_or_none()
        if not schema:
            raise NotFoundError
        return schema

    async def _save(self, payload: dict[str, Any]) -> ConcreteTable:
        schema = self.schema_class(**payload)
        self._session.add(schema)
        await self._session.flush()
        await self._session.refresh(schema)
        return schema

    async def _all(self) -> AsyncGenerator[ConcreteTable, None]:
        result: Result = await self.execute(select(self.schema_class))
        for schema in result.scalars().all():
            yield schema

    async def delete(self, id_: int) -> None:
        await self.execute(
            delete(self.schema_class).where(self.schema_class.id == id_)
        )
        await self._session.flush()


class UsersRepository(BaseRepository[UsersTable]):
    async def all(self) -> AsyncGenerator[UserFlat, None]:
        async for instance in self._all():
            yield UserFlat.model_validate(instance)

    async def get(self, id_: int) -> UserFlat:
        instance = await self._get(key="id", value=id_)
        return UserFlat.model_validate(instance)

    async def get_by_login(self, login: str) -> UserFlat:
        for field in ("username", "email"):
            try:
                instance = await self._get(key=field, value=login)
                return UserFlat.model_validate(instance)
            except NotFoundError:
                continue
        raise NotFoundError

    async def create(self, schema: UserUncommitted) -> UserFlat:
        instance = await self._save(schema.model_dump())
        return UserFlat.model_validate(instance)

    async def update(
        self, attr: str, value: Any, payload: dict[str, Any]
    ) -> UserFlat:
        schema = await self._update(attr, value, payload)
        return UserFlat.model_validate(schema)



from enum import StrEnum
from typing import Any, Mapping

from pydantic import Field, conlist

from .base import PublicEntity

__all__ = (
    "ErrorType",
    "ErrorDetail",
    "ErrorResponse",
    "ErrorResponseMulti",
)


class ErrorType(StrEnum):
    INTERNAL = "internal"
    EXTERNAL = "external"
    VALIDATION = "validation"
    MISSING = "missing"


class ErrorDetail(PublicEntity):
    path: list[str] = Field(default_factory=list)
    type: ErrorType = Field(default=ErrorType.INTERNAL)


class ErrorResponse(PublicEntity):
    message: str
    detail: ErrorDetail = Field(default_factory=ErrorDetail)


class ErrorResponseMulti(PublicEntity):
    result: conlist(ErrorResponse, min_length=1)  # type: ignore[valid-type]


_Response = Mapping[int | str, dict[str, Any]]



"""HTTP helpers for Robyn routes."""

from typing import Any

import msgspec
from robyn import Response
from starlette import status

from .entities.response import ErrorResponse
from .errors import BaseError

JSON_HEADERS = {"content-type": "application/json; charset=utf-8"}


def _encode(payload: Any) -> bytes:
    return msgspec.json.encode(payload)


def json_response(
    payload: Any, status_code: int = status.HTTP_200_OK
) -> Response:
    return Response(
        status_code,
        JSON_HEADERS,
        _encode({"result": payload}),
    )


def error_response(exc: Exception) -> Response:
    if isinstance(exc, BaseError):
        payload = ErrorResponse(message=exc.message).model_dump(by_alias=True)
        body = _encode({"error": payload})
        return Response(exc.status_code, JSON_HEADERS, body)

    payload = ErrorResponse(
        message=str(exc) or "Internal Server Error"
    ).model_dump(by_alias=True)
    body = _encode({"error": payload})
    return Response(status.HTTP_500_INTERNAL_SERVER_ERROR, JSON_HEADERS, body)



"""CORS middleware registration for Robyn."""

from typing import Sequence

from robyn import ALLOW_CORS, Robyn

from ....config import settings


def _normalize_sequence(values: Sequence[object] | None) -> list[str]:
    if values is None:
        return []
    return [str(value) for value in values]


def register(app: Robyn) -> None:
    origins_config = _normalize_sequence(settings.cors.allow_origins)
    headers_config = _normalize_sequence(settings.cors.allow_headers)

    origins: list[str] | str = origins_config or "*"
    if not headers_config or headers_config == ["*"]:
        headers: list[str] | str = "*"
    else:
        headers = headers_config

    # Robyn's helper handles OPTIONS preflight + response headers for us.
    ALLOW_CORS(app, origins=origins, headers=headers)



from datetime import datetime

from pydantic import BaseModel, ConfigDict, EmailStr, field_validator
from pydantic.alias_generators import to_camel

from .authentication import AuthProvider
from .constants import Role


class InternalEntity(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
        validate_assignment=True,
        arbitrary_types_allowed=True,
        from_attributes=True,
    )


class PublicEntity(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        use_enum_values=True,
        validate_assignment=True,
        populate_by_name=True,
        arbitrary_types_allowed=True,
        from_attributes=True,
        loc_by_alias=True,
        alias_generator=to_camel,
    )


class TimeStampMixin(BaseModel):
    created_at: datetime
    updated_at: datetime


class UserUncommitted(InternalEntity):
    username: str
    email: EmailStr
    password: str
    role: int
    is_active: bool = False
    auth_provider: AuthProvider = AuthProvider.INTERNAL

    @field_validator("role", mode="before")
    @classmethod
    def role_validator(cls, value: int) -> int:
        if value not in Role.values():
            raise ValueError(f"Unsupported role: {value}")
        return value


class UserFlat(UserUncommitted, TimeStampMixin):
    id: int


class PasswordForgot(InternalEntity):
    email: EmailStr


class EmailChange(InternalEntity):
    user_id: int
    email: EmailStr



from __future__ import annotations

import json
from datetime import datetime
from typing import Any, Generic, TypeVar, get_args

from pydantic import Field
from redis.asyncio import Redis

from .config import settings
from .schemas import InternalEntity
from .utils import NotFoundError

try:  # Optional dependency for fakeredis in dev
    from fakeredis.aioredis import FakeRedis, FakeServer  # type: ignore
except Exception:  # pragma: no cover - fakeredis not installed
    FakeRedis = None  # type: ignore
    FakeServer = None  # type: ignore

CacheClient = Redis

_fake_server: "FakeServer | None" = None
_CacheEntryInstance = TypeVar("_CacheEntryInstance", bound=InternalEntity)


class CacheEntry(InternalEntity, Generic[_CacheEntryInstance]):
    instance: _CacheEntryInstance
    created_at: datetime = Field(default_factory=datetime.utcnow)


class CacheRepository(Generic[_CacheEntryInstance]):
    def __init__(self) -> None:
        self.redis_client: CacheClient | None = None

    async def __aenter__(self) -> "CacheRepository[_CacheEntryInstance]":
        if settings.cache.use_fake:
            if (
                not FakeRedis or not FakeServer
            ):  # pragma: no cover - import guard
                raise RuntimeError(
                    "fakeredis must be installed to use the fake cache "
                    "(SETTINGS__CACHE__USE_FAKE=true). Install the `[dev]` extras "
                    "or disable the flag."
                )
            global _fake_server
            if _fake_server is None:
                _fake_server = FakeServer()
            self.redis_client = FakeRedis(server=_fake_server)
        else:
            self.redis_client = Redis(
                host=settings.cache.host,
                port=settings.cache.port,
                db=settings.cache.db,
            )
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        if self.redis_client is not None:
            await self.redis_client.close()
        self.redis_client = None

    def _build_key(self, namespace: str, key: Any) -> str:
        return f"{namespace}:{key}"

    def _model_for_entry(self) -> type[_CacheEntryInstance]:
        try:
            return get_args(self.__orig_class__)[0]  # type: ignore[attr-defined]
        except Exception:  # pragma: no cover - fallback
            return InternalEntity  # type: ignore[return-value]

    async def get(
        self, namespace: str, key: Any
    ) -> CacheEntry[_CacheEntryInstance]:
        if self.redis_client is None:
            raise RuntimeError("CacheRepository not initialized")

        built_key = self._build_key(namespace, key)
        raw = await self.redis_client.get(built_key)
        if raw is None:
            raise NotFoundError(message=f"Cache entry not found. Key: {key}")

        try:
            payload = json.loads(raw)
        except (TypeError, json.JSONDecodeError) as exc:  # pragma: no cover
            raise NotFoundError(
                message=f"Cache entry invalid. Key: {key}"
            ) from exc

        model = self._model_for_entry()
        return CacheEntry[_CacheEntryInstance](
            instance=model(**payload["instance"])
        )

    async def set(
        self,
        *,
        namespace: str,
        key: Any,
        instance: _CacheEntryInstance,
        ttl: int | None = None,
    ) -> CacheEntry[_CacheEntryInstance]:
        if self.redis_client is None:
            raise RuntimeError("CacheRepository not initialized")

        entry = CacheEntry[_CacheEntryInstance](instance=instance)
        await self.redis_client.set(
            name=self._build_key(namespace, key),
            value=entry.model_dump_json(),
            ex=ttl,
        )
        return entry

    async def delete(self, namespace: str, key: Any) -> None:
        if self.redis_client is None:
            raise RuntimeError("CacheRepository not initialized")
        await self.redis_client.delete(self._build_key(namespace, key))



from robyn import Request, Response, Robyn

from ..utils import json_response


def register(app: Robyn) -> None:
    @app.get("/health", const=True)
    async def health(_: Request) -> Response:
        return json_response({"status": "ok"})



from typing import Sequence

from pydantic import EmailStr

from ..application import InternalEntity


class EmailMessage(InternalEntity):
    recipients: Sequence[EmailStr]
    subject: str
    body: str



from pydantic import BaseModel


class TokenSettings(BaseModel):
    secret_key: str = "dev-secret"
    ttl: int = 3600


class Settings(BaseModel):
    algorithm: str = "HS256"
    scheme: str = "Bearer"
    access_token: TokenSettings = TokenSettings(ttl=3600)
    refresh_token: TokenSettings = TokenSettings(ttl=86400)
    user_activation_ttl: int = 0
    password_reset_token_ttl: int = 0
    session_secret_key: str = "change-me"



from datetime import datetime
from typing import TypeVar

from sqlalchemy import MetaData, SmallInteger, func
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from ..authentication import AuthProvider

__all__ = (
    "Base",
    "ConcreteTable",
    "UsersTable",
    "NewsLetterSubscriptionsTable",
)

meta = MetaData(
    naming_convention={
        "ix": "ix_%(column_0_label)s",
        "uq": "uq_%(table_name)s_%(column_0_name)s",
        "ck": "ck_%(table_name)s_%(constraint_name)s",
        "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
        "pk": "pk_%(table_name)s",
    }
)


class Base(DeclarativeBase):
    metadata = meta


class TimestampMixin:
    id: Mapped[int] = mapped_column(primary_key=True)
    created_at: Mapped[datetime] = mapped_column(
        default=datetime.utcnow, server_default=func.CURRENT_TIMESTAMP()
    )
    updated_at: Mapped[datetime] = mapped_column(
        default=datetime.utcnow,
        onupdate=datetime.utcnow,
        server_default=func.CURRENT_TIMESTAMP(),
    )


class BaseTable(TimestampMixin, Base):
    __abstract__ = True


ConcreteTable = TypeVar("ConcreteTable", bound="BaseTable")


class UsersTable(BaseTable):
    __tablename__ = "users"

    username: Mapped[str] = mapped_column(unique=True, index=True)
    email: Mapped[str] = mapped_column(unique=True, index=True)
    password: Mapped[str | None] = mapped_column()
    is_active: Mapped[bool] = mapped_column(default=False, server_default="0")
    role: Mapped[int] = mapped_column(
        SmallInteger, default=1, server_default="1"
    )
    auth_provider: Mapped[str] = mapped_column(
        default=AuthProvider.INTERNAL,
        server_default=AuthProvider.INTERNAL,
    )


class NewsLetterSubscriptionsTable(BaseTable):
    __tablename__ = "news_letter_subscriptions"

    email: Mapped[str] = mapped_column(unique=True, index=True)



"""Create module for Robyn project scaffolding."""

from .utils import (
    DESIGN_CHOICES,
    ORM_CHOICES,
    PACKAGE_MANAGER_CHOICES,
    apply_package_manager,
    collect_existing_items,
    copy_template,
    ensure_package_manager_available,
    get_generated_items,
    prepare_destination,
)

__all__ = [
    "ORM_CHOICES",
    "DESIGN_CHOICES",
    "PACKAGE_MANAGER_CHOICES",
    "ensure_package_manager_available",
    "collect_existing_items",
    "get_generated_items",
    "prepare_destination",
    "copy_template",
    "apply_package_manager",
]



"""Input validators reused by presentation layers."""

from ...infrastructure.application import UnprocessableError


def password_nist(payload: str) -> str:
    """Check if the password suits the NIST requirements."""

    if len(payload) < 8:
        raise UnprocessableError(
            message="The password must be at least 8 characters long."
        )

    if len(payload) > 64:
        raise UnprocessableError(
            message="The password must be at most 64 characters long."
        )

    if not any(char.isdigit() for char in payload):
        raise UnprocessableError(
            message="The password must contain at least one digit."
        )

    if not any(char.isupper() for char in payload):
        raise UnprocessableError(
            message="The password must contain at least one uppercase letter."
        )

    if not any(char.islower() for char in payload):
        raise UnprocessableError(
            message="The password must contain at least one lowercase letter."
        )

    if not any(char in "!@#$%^&*()-+?_=,<>/;:[]{}" for char in payload):
        raise UnprocessableError(
            message="The password must contain at least one special character."
        )

    return payload



"""Authentication workflows: verifying credentials and handling tokens."""

from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import Any, Dict

import jwt
from robyn import Request
from robyn.authentication import AuthenticationHandler, BearerGetter, Identity

from ..config import settings
from ..domain.users import UserFlat
from ..infrastructure.application import (
    AuthenticationError,
    NotFoundError,
    json_response,
)
from ..infrastructure.application.entities.response import ErrorResponse
from ..infrastructure.authentication import pwd_context
from ..infrastructure.database import transaction
from ..infrastructure.database.repository import (
    UsersRepository as InfrastructureUsersRepository,
)


async def authenticate_user(login: str, password: str) -> UserFlat:
    """Validate credentials and return the matching active user."""
    try:
        async with transaction():
            repository = InfrastructureUsersRepository()
            user = await repository.get_by_login(login=login)
    except NotFoundError as exc:
        raise AuthenticationError(message="Invalid credentials") from exc

    if not pwd_context.verify(password, user.password, scheme="bcrypt"):
        raise AuthenticationError(message="Invalid credentials")

    if not user.is_active:
        raise AuthenticationError(message="Inactive account")

    return user


def create_access_token(user: UserFlat) -> str:
    """Generate a signed JWT for the provided user."""
    now = datetime.now(timezone.utc)
    ttl = settings.authentication.access_token.ttl
    exp = now + timedelta(seconds=ttl)
    payload = {
        "sub": str(user.id),
        "email": user.email,
        "username": user.username,
        "iat": int(now.timestamp()),
        "exp": int(exp.timestamp()),
    }
    return jwt.encode(
        payload,
        settings.authentication.access_token.secret_key,
        algorithm=settings.authentication.algorithm,
    )


def decode_access_token(token: str) -> Dict[str, Any]:
    """Decode a JWT and return its claims, raising AuthenticationError on failure."""
    try:
        payload = jwt.decode(
            token,
            settings.authentication.access_token.secret_key,
            algorithms=[settings.authentication.algorithm],
        )
    except jwt.ExpiredSignatureError as exc:
        raise AuthenticationError(message="Token has expired") from exc
    except jwt.PyJWTError as exc:
        raise AuthenticationError(message="Invalid token") from exc

    return payload


def require_user_id(request: Request) -> int:
    identity: Identity | None = getattr(request, "identity", None)
    if identity is None:
        raise AuthenticationError(message="Authentication required")

    sub: str | Any | None = identity.claims.get("sub")
    if sub is None:
        raise AuthenticationError(message="Subject claim missing")

    try:
        return int(sub)
    except (TypeError, ValueError) as exc:
        raise AuthenticationError(message="Subject claim invalid") from exc


class JWTAuthenticationHandler(AuthenticationHandler):
    """Robyn authentication handler that validates bearer tokens."""

    def __init__(self) -> None:
        super().__init__(BearerGetter())
        self._last_error = AuthenticationError()

    def authenticate(self, request) -> Identity | None:  # type: ignore[override]
        token = self.token_getter.get_token(request)
        if not token:
            self._last_error = AuthenticationError(
                message="Token not provided"
            )
            return None

        try:
            claims = decode_access_token(token)
        except AuthenticationError as exc:
            self._last_error = exc
            return None

        normalized = {str(key): str(value) for key, value in claims.items()}
        return Identity(claims=normalized)

    @property
    def unauthorized_response(self):
        payload = ErrorResponse(message=self._last_error.message).model_dump(
            by_alias=True
        )
        return json_response(payload, status_code=self._last_error.status_code)



"""Runtime entrypoint for production images.

Runs Alembic migrations before handing off to the Robyn server (or any custom command).
"""

from __future__ import annotations

import os
import shlex
import subprocess
import sys


def _run(cmd: list[str]) -> None:
    subprocess.run(cmd, check=True)


def main() -> None:
    _run(["alembic", "upgrade", "head"])

    raw_cmd = os.environ.get("APP_CMD")
    if raw_cmd:
        _run(shlex.split(raw_cmd))
        return

    _run(
        [
            sys.executable,
            "-m",
            "app.server",
            "--fast",
            "--processes",
            "4",
            "--workers",
            "3",
            "--log-level",
            "WARNING",
        ]
    )


if __name__ == "__main__":
    main()



import uuid

from robyn import Request, Response, Robyn
from starlette import status

from ..authentication import AuthProvider, pwd_context
from ..cache import CacheRepository
from ..config import settings
from ..mailing import EmailMessage, mailing_service
from ..models import UsersRepository, transaction
from ..schemas import EmailChange, UserFlat, UserUncommitted
from ..utils import (
    DatabaseError,
    NotFoundError,
    UnprocessableError,
    json_response,
)
from .authentication import require_user_id
from .contracts import (
    ActivationBody,
    EmailChangeConfirmBody,
    EmailChangeRequestBody,
    PasswordChangeBody,
    PasswordResetConfirmBody,
    PasswordResetRequestBody,
    UserCreateBody,
    UserPublic,
)
from .helpers import parse_body


def _serialize(user: UserFlat) -> dict:
    return UserPublic.model_validate(user).model_dump(by_alias=True)


def _activation_link(key: uuid.UUID) -> str:
    base = str(settings.integrations.frontend.activation_base_url).rstrip("/")
    return f"{base}/{key}"


def _password_reset_link(key: uuid.UUID) -> str:
    base = str(settings.integrations.frontend.password_reset_base_url).rstrip(
        "/"
    )
    return f"{base}/{key}"


def _email_change_link(key: uuid.UUID) -> str:
    base = str(settings.integrations.frontend.email_change_base_url).rstrip(
        "/"
    )
    return f"{base}/{key}"


async def _cache_activation(user: UserFlat, key: uuid.UUID) -> None:
    async with CacheRepository[UserFlat]() as cache:
        await cache.set(
            namespace="activation",
            key=key,
            instance=user,
            ttl=settings.cache.ttl_activation_seconds or None,
        )


async def _send_activation_email(user: UserFlat, key: uuid.UUID) -> None:
    link = _activation_link(key)
    await mailing_service.send(
        EmailMessage(
            recipients=[user.email],
            subject="Account activation",
            body=f"Activate your account: {link}",
        )
    )


async def _cache_password_reset(user: UserFlat, key: uuid.UUID) -> None:
    async with CacheRepository[UserFlat]() as cache:
        await cache.set(
            namespace="password-reset",
            key=key,
            instance=user,
            ttl=settings.cache.ttl_password_reset_seconds or None,
        )


async def _send_password_reset_email(user: UserFlat, key: uuid.UUID) -> None:
    link = _password_reset_link(key)
    await mailing_service.send(
        EmailMessage(
            recipients=[user.email],
            subject="Password reset",
            body=f"Reset your password: {link}",
        )
    )


async def _cache_email_change(entry: EmailChange, key: uuid.UUID) -> None:
    async with CacheRepository[EmailChange]() as cache:
        await cache.set(
            namespace="email-change",
            key=key,
            instance=entry,
        )


async def _send_email_change_email(email: str, key: uuid.UUID) -> None:
    link = _email_change_link(key)
    await mailing_service.send(
        EmailMessage(
            recipients=[email],
            subject="Confirm email change",
            body=f"Confirm your email change: {link}",
        )
    )


async def _create_user(payload: UserCreateBody) -> UserFlat:
    async with transaction():
        repo = UsersRepository()
        schema = UserUncommitted(
            username=payload.username,
            email=payload.email,
            password=pwd_context.hash(payload.password, scheme="bcrypt"),
            role=1,
        )
        try:
            user = await repo.create(schema)
        except DatabaseError as exc:
            raise UnprocessableError(message="User already exists") from exc

    activation_key = uuid.uuid4()
    await _cache_activation(user, activation_key)
    await _send_activation_email(user, activation_key)
    return user


def register(app: Robyn) -> None:
    @app.post("/users")
    async def user_create(request: Request) -> Response:
        payload = await parse_body(request, UserCreateBody)
        user = await _create_user(payload)
        return json_response(
            payload=_serialize(user),
            status_code=status.HTTP_201_CREATED,
        )

    @app.post("/users/activate")
    async def user_activate(request: Request) -> Response:
        payload = await parse_body(request, ActivationBody)

        async with CacheRepository[UserFlat]() as cache:
            try:
                cache_entry = await cache.get(
                    namespace="activation", key=payload.key
                )
            except NotFoundError as exc:
                raise UnprocessableError(
                    message="Invalid or expired activation key"
                ) from exc

            async with transaction():
                repo = UsersRepository()
                user = await repo.update(
                    attr="id",
                    value=cache_entry.instance.id,
                    payload={"is_active": True},
                )

            await cache.delete(namespace="activation", key=payload.key)

        return json_response(
            payload=_serialize(user),
            status_code=status.HTTP_200_OK,
        )

    @app.post("/users/password/change", auth_required=True)
    async def user_password_change(request: Request) -> Response:
        payload = await parse_body(request, PasswordChangeBody)
        user_id = require_user_id(request)

        async with transaction():
            repo = UsersRepository()
            user = await repo.get(id_=user_id)

            if not pwd_context.verify(
                payload.old_password, user.password, scheme="bcrypt"
            ):
                raise UnprocessableError(message="Password invalid")

            updated = await repo.update(
                attr="id",
                value=user.id,
                payload={
                    "password": pwd_context.hash(
                        payload.new_password, scheme="bcrypt"
                    ),
                    "auth_provider": AuthProvider.INTERNAL,
                },
            )

        return json_response(
            payload=_serialize(updated),
            status_code=status.HTTP_200_OK,
        )

    @app.post("/users/password/reset/request")
    async def user_password_reset_request(request: Request) -> Response:
        payload = await parse_body(request, PasswordResetRequestBody)

        try:
            async with transaction():
                repo = UsersRepository()
                user = await repo.get_by_login(login=payload.email)
        except NotFoundError:
            return json_response({}, status_code=status.HTTP_202_ACCEPTED)

        reset_key = uuid.uuid4()
        await _cache_password_reset(user, reset_key)
        await _send_password_reset_email(user, reset_key)
        return json_response({}, status_code=status.HTTP_202_ACCEPTED)

    @app.post("/users/password/reset/confirm")
    async def user_password_reset_confirm(request: Request) -> Response:
        payload = await parse_body(request, PasswordResetConfirmBody)

        async with CacheRepository[UserFlat]() as cache:
            try:
                cache_entry = await cache.get(
                    namespace="password-reset", key=payload.key
                )
            except NotFoundError as exc:
                raise UnprocessableError(
                    message="Invalid or expired reset key"
                ) from exc

            async with transaction():
                repo = UsersRepository()
                user = await repo.update(
                    attr="id",
                    value=cache_entry.instance.id,
                    payload={
                        "password": pwd_context.hash(
                            payload.password, scheme="bcrypt"
                        ),
                        "auth_provider": AuthProvider.INTERNAL,
                    },
                )

            await cache.delete(namespace="password-reset", key=payload.key)

        return json_response(
            payload=_serialize(user),
            status_code=status.HTTP_200_OK,
        )

    @app.post("/users/email-change/request", auth_required=True)
    async def user_email_change_request(request: Request) -> Response:
        payload = await parse_body(request, EmailChangeRequestBody)
        user_id = require_user_id(request)

        async with transaction():
            repo = UsersRepository()
            user = await repo.get(id_=user_id)

        change_key = uuid.uuid4()
        entry = EmailChange(user_id=user.id, email=payload.email)
        await _cache_email_change(entry, change_key)
        await _send_email_change_email(payload.email, change_key)

        return json_response({}, status_code=status.HTTP_202_ACCEPTED)

    @app.post("/users/email-change/confirm")
    async def user_email_change_confirm(request: Request) -> Response:
        payload = await parse_body(request, EmailChangeConfirmBody)

        async with CacheRepository[EmailChange]() as cache:
            try:
                cache_entry = await cache.get(
                    namespace="email-change", key=payload.key
                )
            except NotFoundError as exc:
                raise UnprocessableError(
                    message="Invalid or expired email change key"
                ) from exc

            async with transaction():
                repo = UsersRepository()
                user = await repo.update(
                    attr="id",
                    value=cache_entry.instance.user_id,
                    payload={
                        "email": cache_entry.instance.email,
                        "auth_provider": AuthProvider.INTERNAL,
                    },
                )

            await cache.delete(namespace="email-change", key=payload.key)

        return json_response(
            payload=_serialize(user),
            status_code=status.HTTP_200_OK,
        )



"""Project scaffolding CLI based on the local Robyn template."""

from __future__ import annotations

import shutil
import tempfile
from pathlib import Path

import click

from add import add_business_logic
from create import (
    DESIGN_CHOICES,
    ORM_CHOICES,
    PACKAGE_MANAGER_CHOICES,
    apply_package_manager,
    collect_existing_items,
    copy_template,
    ensure_package_manager_available,
    get_generated_items,
    prepare_destination,
)


def _remove_path(path: Path) -> None:
    if path.is_dir():
        shutil.rmtree(path, ignore_errors=True)
    else:
        path.unlink(missing_ok=True)


def _cleanup_create_failure(
    target_dir: Path,
    generated_items: set[Path],
    existing_items: set[Path],
    created_new_dir: bool,
) -> None:
    """Attempt to remove files created during a failed create command."""
    if created_new_dir and target_dir.exists():
        shutil.rmtree(target_dir, ignore_errors=True)
        return

    for rel_path in generated_items:
        if rel_path in existing_items:
            continue
        candidate = target_dir / rel_path
        if candidate.exists():
            _remove_path(candidate)


def _backup_project(project_path: Path) -> tuple[Path, Path]:
    """Create a backup of the project directory for rollback."""
    temp_dir = Path(tempfile.mkdtemp(prefix="robyn-config-add-backup-"))
    backup_path = temp_dir / "project"
    shutil.copytree(project_path, backup_path, dirs_exist_ok=True)
    return temp_dir, backup_path


def _restore_project_backup(project_path: Path, backup_path: Path) -> None:
    """Restore project directory from backup."""
    if project_path.exists():
        for child in project_path.iterdir():
            _remove_path(child)
    shutil.copytree(backup_path, project_path, dirs_exist_ok=True)


@click.group(name="robyn-config")
def cli() -> None:
    """Robyn configuration utilities."""


@cli.command("create")
@click.argument("name")
@click.option(
    "-orm",
    "--orm",
    "orm_type",
    type=click.Choice(ORM_CHOICES, case_sensitive=False),
    default="sqlalchemy",
    show_default=True,
    help="Select the ORM implementation to copy.",
)
@click.option(
    "-design",
    "--design",
    "design",
    type=click.Choice(DESIGN_CHOICES, case_sensitive=False),
    default="ddd",
    show_default=True,
    help="Select the design pattern.",
)
@click.option(
    "-package-manager",
    "--package-manager",
    "package_manager",
    type=click.Choice(PACKAGE_MANAGER_CHOICES, case_sensitive=False),
    default="uv",
    show_default=True,
    help="Select the package manager to use.",
)
@click.argument(
    "destination",
    type=click.Path(
        exists=False, file_okay=False, dir_okay=True, path_type=Path
    ),
    default=".",
)
def create(
    name: str,
    destination: Path,
    orm_type: str,
    design: str,
    package_manager: str,
) -> None:
    """Copy the template into destination with specific configurations."""
    package_manager = package_manager.lower()
    ensure_package_manager_available(package_manager)

    destination_resolved = destination.expanduser().resolve()
    destination_exists_before = destination_resolved.exists()
    existing_items: set[Path] = set()
    if destination_exists_before:
        existing_items = collect_existing_items(destination_resolved)

    target_dir: Path | None = None
    generated_items: set[Path] = set()
    created_new_dir = False

    try:
        click.echo(f"Creating Robyn template ({design}/{orm_type})...")
        target_dir = prepare_destination(
            destination, orm_type, design, package_manager
        )
        generated_items = get_generated_items(
            orm_type, design, package_manager
        )
        created_new_dir = not destination_exists_before and target_dir.exists()

        copy_template(target_dir, orm_type, design, name, package_manager)

        click.echo("Installing dependencies...")
        apply_package_manager(target_dir, package_manager)

        click.echo(
            click.style("Successfully created Robyn template", fg="green")
        )
    except Exception as e:
        if target_dir:
            _cleanup_create_failure(
                target_dir, generated_items, existing_items, created_new_dir
            )
        raise click.ClickException(click.style(str(e), fg="red")) from e


@cli.command("add")
@click.argument("name")
@click.argument(
    "project_path",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, path_type=Path
    ),
    default=".",
)
def add(name: str, project_path: Path) -> None:
    """Add new business logic to an existing robyn-config project."""
    backup_dir: Path | None = None
    backup_path: Path | None = None
    project_path = project_path.resolve()

    try:
        backup_dir, backup_path = _backup_project(project_path)
        add_business_logic(project_path, name)
        click.echo(
            click.style(
                f"Successfully added '{name}' business logic!", fg="green"
            )
        )
    except Exception as e:
        if backup_path:
            _restore_project_backup(project_path, backup_path)
        raise click.ClickException(click.style(str(e), fg="red")) from e
    finally:
        if backup_dir:
            shutil.rmtree(backup_dir, ignore_errors=True)


if __name__ == "__main__":
    cli()



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Generic, TypeVar, get_args

from redis.asyncio import Redis
from redis.asyncio.client import Pipeline

from ...config import settings
from ..application import InternalEntity, NotFoundError
from .entities import CacheEntry

if TYPE_CHECKING:
    from fakeredis.aioredis import FakeRedis  # pragma: no cover
    from fakeredis.aioredis import FakeServer  # pragma: no cover

    CacheClient = Redis | FakeRedis
else:
    CacheClient = Redis  # type: ignore[assignment]

_fake_server: "FakeServer" | None = None

_CacheEntryInstance = TypeVar("_CacheEntryInstance", bound=InternalEntity)


class CacheRepository(Generic[_CacheEntryInstance]):
    def __init__(self) -> None:
        self.redis_client: CacheClient | None = None
        self.transaction: Pipeline | None = None

    async def __aenter__(self) -> "CacheRepository[_CacheEntryInstance]":
        if settings.cache.use_fake:
            try:
                from fakeredis.aioredis import FakeRedis, FakeServer
            except ImportError as exc:
                raise RuntimeError(
                    "fakeredis must be installed to use the fake cache (SETTINGS__CACHE__USE_FAKE=true);"
                    " install the `[dev]` extras or disable the flag."
                ) from exc
            global _fake_server
            if _fake_server is None:
                _fake_server = FakeServer()
            self.redis_client = FakeRedis(server=_fake_server)
        else:
            self.redis_client = Redis(
                host=settings.cache.host,
                port=settings.cache.port,
                db=settings.cache.db,
            )

        assert self.redis_client is not None
        self.transaction = self.redis_client.pipeline(transaction=True)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        assert self.redis_client is not None
        assert self.transaction is not None
        if exc_type is None:
            await self.transaction.execute()
        await self.redis_client.close()
        self.redis_client = None
        self.transaction = None

    def _build_key(self, namespace: str, key: Any) -> str:
        return f"{namespace}:{key}"

    async def get(
        self, namespace: str, key: Any
    ) -> CacheEntry[_CacheEntryInstance]:
        assert self.transaction is not None
        built_key = self._build_key(namespace, key)
        results = await self.transaction.get(built_key).execute()  # type: ignore[union-attr]
        if not results or results[0] is None:
            raise NotFoundError(message=f"Cache entry not found. Key: {key}")

        try:
            payload = json.loads(results[0])
        except (TypeError, json.JSONDecodeError) as exc:  # pragma: no cover
            raise NotFoundError(
                message=f"Cache entry invalid. Key: {key}"
            ) from exc

        struct = get_args(self.__orig_class__)[0]  # type: ignore[attr-defined]
        return CacheEntry[_CacheEntryInstance](
            instance=struct(**payload["instance"])
        )

    async def set(
        self,
        *,
        namespace: str,
        key: Any,
        instance: _CacheEntryInstance,
        ttl: int | None = None,
    ) -> CacheEntry[_CacheEntryInstance]:
        assert self.transaction is not None
        entry = CacheEntry[_CacheEntryInstance](instance=instance)
        await self.transaction.set(
            name=self._build_key(namespace, key),
            value=entry.model_dump_json(),
            ex=ttl,
        ).execute()  # type: ignore[union-attr]
        return entry

    async def delete(self, namespace: str, key: Any) -> None:
        assert self.transaction is not None
        await self.transaction.delete(self._build_key(namespace, key)).execute()  # type: ignore[union-attr]



"""Authentication routes: login and token introspection."""

from datetime import datetime, timezone

from robyn import Request, Response, Robyn
from robyn.authentication import Identity
from starlette import status

from ...infrastructure.application import AuthenticationError, json_response
from ...operational import authentication as auth_ops
from .._helpers import parse_body
from .contracts import LoginRequestBody, TokenInfo, TokenResponse


def register(app: Robyn) -> None:
    @app.post("/auth/login")
    async def login(request: Request) -> Response:
        payload = await parse_body(request, LoginRequestBody)
        user = await auth_ops.authenticate_user(
            payload.login, payload.password
        )
        token = auth_ops.create_access_token(user)
        response = TokenResponse(access_token=token)
        return json_response(
            payload=response.model_dump(by_alias=True),
            status_code=status.HTTP_200_OK,
        )

    @app.get("/auth/me", auth_required=True)
    async def me(request: Request) -> Response:
        identity: Identity | None = getattr(request, "identity", None)
        if identity is None:
            raise AuthenticationError(message="Authentication required")

        claims = identity.claims
        issued_at = datetime.fromtimestamp(
            float(claims.get("iat", "0")), tz=timezone.utc
        )
        expires_at = datetime.fromtimestamp(
            float(claims.get("exp", "0")), tz=timezone.utc
        )
        info = TokenInfo(
            subject=claims.get("sub", ""),
            email=claims.get("email", ""),
            username=claims.get("username", ""),
            issued_at=issued_at,
            expires_at=expires_at,
        )
        return json_response(
            payload=info.model_dump(by_alias=True),
            status_code=status.HTTP_200_OK,
        )



from . import middlewares  # noqa: F401
from .entities import *  # noqa: F401, F403
from .errors import *  # noqa: F401, F403
from .factory import create  # noqa: F401
from .http import error_response, json_response  # noqa: F401



import logging
from pathlib import Path

from app.config import settings
from app.infrastructure.application import error_response, middlewares
from app.infrastructure.application.factory import create
from app.operational.authentication import JWTAuthenticationHandler
from app.presentation import register_routes
from loguru import logger
from robyn import Robyn

log_path = Path(settings.root_dir) / "logs"
log_path.mkdir(exist_ok=True)
logger.add(
    log_path / f"{settings.logging.file}.log",
    format=settings.logging.format,
    rotation=settings.logging.rotation,
    compression=settings.logging.compression,
    level="INFO",
)

if not settings.debug:
    logging.getLogger("sqlalchemy.engine").setLevel(logging.WARNING)
    logging.getLogger("alembic").setLevel(logging.WARNING)

app: Robyn = create(
    __name__,
    route_registrars=(register_routes,),
    middlewares=(
        middlewares.sessions.register,
        middlewares.cors.register,
    ),
    exception_handler=error_response,
    authentication_handler=JWTAuthenticationHandler(),
)


if __name__ == "__main__":
    app.start(host="0.0.0.0", port=8000)



from typing import Any, AsyncGenerator, Generic, get_args, get_origin

from tortoise.expressions import Q
from tortoise.queryset import QuerySet

from ..schemas import UserFlat, UserUncommitted
from ..utils import DatabaseError, NotFoundError, UnprocessableError
from .database import Session
from .models import ConcreteTable, UsersTable


class BaseRepository(Session, Generic[ConcreteTable]):
    schema_class: type[ConcreteTable] | None = None

    def __init__(self) -> None:
        super().__init__()
        if not getattr(self, "schema_class", None):
            raise UnprocessableError(message="schema_class is required")

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if cls is BaseRepository:
            return
        if cls.schema_class is None:
            for base in getattr(cls, "__orig_bases__", ()):
                if get_origin(base) is BaseRepository:
                    args = get_args(base)
                    if args:
                        cls.schema_class = args[0]
                        break
        if cls.schema_class is None:
            raise UnprocessableError(message="schema_class is required")

    def _query(self) -> QuerySet[ConcreteTable]:
        return self.schema_class.all().using_db(self._connection)

    def _filter(
        self, *expressions: Q, **filters: Any
    ) -> QuerySet[ConcreteTable]:
        return self.schema_class.filter(*expressions, **filters).using_db(
            self._connection
        )

    async def _update(
        self, key: str, value: Any, payload: dict[str, Any]
    ) -> ConcreteTable:
        filters = {key: value}
        updated = await self._filter(**filters).update(**payload)
        if not updated:
            raise DatabaseError
        schema = await self._filter(**filters).first()
        if not schema:
            raise DatabaseError
        return schema

    async def _get(self, key: str, value: Any) -> ConcreteTable:
        schema = await self._filter(**{key: value}).first()
        if not schema:
            raise NotFoundError
        return schema

    async def count(self) -> int:
        value = await self._query().count()
        if not isinstance(value, int):
            raise UnprocessableError(message=f"Count returned {value}")
        return value

    async def _first(self, by: str = "id") -> ConcreteTable:
        schema = await self._query().order_by(by).first()
        if not schema:
            raise NotFoundError
        return schema

    async def _last(self, by: str = "id") -> ConcreteTable:
        schema = await self._query().order_by(f"-{by}").first()
        if not schema:
            raise NotFoundError
        return schema

    async def _save(self, payload: dict[str, Any]) -> ConcreteTable:
        schema = self.schema_class(**payload)
        await schema.save(using_db=self._connection)
        await schema.refresh_from_db(using_db=self._connection)
        return schema

    async def _all(self) -> AsyncGenerator[ConcreteTable, None]:
        async for schema in self._query():
            yield schema

    async def delete(self, id_: int) -> None:
        await self._filter(id=id_).delete()


class UsersRepository(BaseRepository[UsersTable]):
    async def all(self) -> AsyncGenerator[UserFlat, None]:
        async for instance in self._all():
            yield UserFlat.model_validate(instance)

    async def get(self, id_: int) -> UserFlat:
        instance = await self._get(key="id", value=id_)
        return UserFlat.model_validate(instance)

    async def get_by_login(self, login: str) -> UserFlat:
        for field in ("username", "email"):
            try:
                instance = await self._get(key=field, value=login)
                return UserFlat.model_validate(instance)
            except NotFoundError:
                continue
        raise NotFoundError

    async def create(self, schema: UserUncommitted) -> UserFlat:
        instance = await self._save(schema.model_dump())
        return UserFlat.model_validate(instance)

    async def update(
        self, attr: str, value: Any, payload: dict[str, Any]
    ) -> UserFlat:
        schema = await self._update(attr, value, payload)
        return UserFlat.model_validate(schema)



"""Base entities used across the Robyn app."""

from typing import TypeVar

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel, to_snake


class InternalEntity(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        use_enum_values=True,
        validate_assignment=True,
        arbitrary_types_allowed=True,
        from_attributes=True,
    )


_InternalEntity = TypeVar("_InternalEntity", bound=InternalEntity)


class PublicEntity(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        use_enum_values=True,
        validate_assignment=True,
        populate_by_name=True,
        arbitrary_types_allowed=True,
        from_attributes=True,
        loc_by_alias=True,
        alias_generator=to_camel,
    )


_PublicEntity = TypeVar("_PublicEntity", bound=PublicEntity)


class PublicSnakeCaseEntity(BaseModel):
    model_config = ConfigDict(
        extra="ignore",
        use_enum_values=True,
        validate_assignment=True,
        populate_by_name=True,
        arbitrary_types_allowed=True,
        from_attributes=True,
        loc_by_alias=True,
        alias_generator=to_snake,
    )


_PublicSnakeCaseEntity = TypeVar(
    "_PublicSnakeCaseEntity", bound=PublicSnakeCaseEntity
)



from datetime import datetime, timedelta

from app.infrastructure.application import PublicEntity
from pydantic import EmailStr, Field


class LoginRequestBody(PublicEntity):
    login: EmailStr | str = Field(
        description="User login", examples=["john@email.com", "john"]
    )
    password: str = Field(description="User password", examples=["password"])


class TokenResponse(PublicEntity):
    access_token: str = Field(description="Access token", examples=["token"])
    token_type: str = Field(
        description="Token type", examples=["Bearer"], default="Bearer"
    )


class TokenInfo(PublicEntity):
    subject: str = Field(description="Subject", examples=["1"])
    email: EmailStr = Field(
        description="User email", examples=["john@email.com"]
    )
    username: str = Field(description="User username", examples=["john"])
    issued_at: datetime = Field(
        description="Issued at", examples=[datetime.now()]
    )
    expires_at: datetime = Field(
        description="Expires at",
        examples=[datetime.now() + timedelta(hours=1)],
    )



"""Utility functions for the 'create' command."""

from __future__ import annotations

import os
import shutil
import subprocess
from pathlib import Path
from typing import Iterable, Mapping, Sequence

import click
from jinja2 import Environment, StrictUndefined

ORM_CHOICES: Sequence[str] = ("sqlalchemy", "tortoise")
DESIGN_CHOICES: Sequence[str] = ("ddd", "mvc")
PACKAGE_MANAGER_CHOICES: Sequence[str] = ("uv", "poetry")
PACKAGE_ROOT = Path(__file__).resolve().parent
SRC_DIR = PACKAGE_ROOT.resolve()
COMMON_DIR = (SRC_DIR / "common").resolve()
COMPOSE_APP_DIR = (COMMON_DIR / "compose" / "app").resolve()
LOCK_FILE_BY_MANAGER: Mapping[str, str] = {
    "uv": "uv.lock",
    "poetry": "poetry.lock",
}
PACKAGE_MANAGER_DOWNLOAD_URLS: Mapping[str, str] = {
    "uv": "https://docs.astral.sh/uv/getting-started/installation/",
    "poetry": "https://python-poetry.org/docs/#installation",
}

TEMPLATE_CONFIGS: Mapping[str, dict[str, str]] = {
    "ddd:sqlalchemy": {
        "design": "ddd",
        "orm": "sqlalchemy",
        "alembic_script_location": "src/app/infrastructure/database/migrations",
    },
    "ddd:tortoise": {
        "design": "ddd",
        "orm": "tortoise",
        "tortoise_orm_path": "src.app.infrastructure.database.services.engine.TORTOISE_ORM",
        "tortoise_migrations_location": "./src/app/infrastructure/database/migrations",
    },
    "mvc:sqlalchemy": {
        "design": "mvc",
        "orm": "sqlalchemy",
        "alembic_script_location": "src/app/models/migrations",
    },
    "mvc:tortoise": {
        "design": "mvc",
        "orm": "tortoise",
        "tortoise_orm_path": "app.models.database.TORTOISE_ORM",
        "tortoise_migrations_location": "./src/app/models/migrations",
    },
}

JINJA_ENV = Environment(undefined=StrictUndefined)


def _collect_existing_items(destination: Path) -> set[Path]:
    items: set[Path] = set()
    for current, dirs, files in os.walk(destination):
        rel_base = Path(current).relative_to(destination)
        for name in dirs:
            items.add(rel_base / name)
        for name in files:
            items.add(rel_base / name)
    return items


def _collect_common_items(orm_type: str, package_manager: str) -> set[Path]:
    items: set[Path] = set()
    for item in COMMON_DIR.iterdir():
        if item.is_dir() or item.name == ".DS_Store":
            continue

        name = item.name
        if item.suffix == ".jinja2":
            name = item.stem

        if orm_type == "tortoise" and name == "alembic.ini":
            continue
        if name in LOCK_FILE_BY_MANAGER.values():
            continue
        items.add(Path(name))
    lock_file = LOCK_FILE_BY_MANAGER.get(package_manager)
    if lock_file:
        items.add(Path(lock_file))
    return items


def _collect_compose_items(_orm_type: str) -> set[Path]:
    items: set[Path] = set()
    target_dir = Path("compose") / "app"

    items.update(
        {
            target_dir / "dev.sh",
            target_dir / "prod.py",
            target_dir / "Dockerfile",
        }
    )

    return items


def _collect_generated_items(
    orm_type: str, design: str, package_manager: str
) -> set[Path]:
    items: set[Path] = set()
    items.update(_collect_common_items(orm_type, package_manager))
    items.update({Path("src")})
    items.update(_collect_compose_items(orm_type))
    return items


def prepare_destination(
    path_arg: str, orm_type: str, design: str, package_manager: str
) -> Path:
    """Prepare and validate the destination directory for the template."""
    destination = Path(path_arg).expanduser().resolve()
    generated_items = _collect_generated_items(
        orm_type, design, package_manager
    )

    if destination.exists():
        if not destination.is_dir():
            print(f"Target path '{destination}' is not a directory.")
            raise SystemExit(1)
        existing_items = _collect_existing_items(destination)
        if existing_items:
            overlapping_items = existing_items & generated_items

            if overlapping_items:
                print(f"Target directory '{destination}' is not empty.")
                print("The following items will be replaced:")
                for item in sorted(overlapping_items):
                    print(f"- {item.as_posix()}")
            else:
                return destination

            proceed = click.confirm(
                "Proceed with applying the template to this directory?",
                default=False,
            )
            if not proceed:
                print("Operation cancelled by user.")
                raise SystemExit(1)
    else:
        destination.mkdir(parents=True, exist_ok=True)
    return destination


def _get_template_config(
    design: str, orm_type: str, project_name: str, package_manager: str
) -> dict[str, str]:
    """Get the template configuration for the given design and ORM type."""
    key = f"{design}:{orm_type}"
    config = TEMPLATE_CONFIGS.get(key)
    if config is None:
        print(
            f"Unsupported configuration '{key}'. "
            f"Valid options: {', '.join(TEMPLATE_CONFIGS.keys())}."
        )
        raise SystemExit(1)
    return {
        **config,
        "name": project_name,
        "package_manager": package_manager,
    }


def _render_template(
    source: Path, target: Path, context: Mapping[str, str]
) -> None:
    """Render a Jinja2 template from source to target."""
    template = JINJA_ENV.from_string(source.read_text())
    rendered = template.render(**context)
    target.parent.mkdir(parents=True, exist_ok=True)
    target.write_text(rendered)


def _copy_common_files(
    destination: Path,
    orm_type: str,
    package_manager: str,
    context: Mapping[str, str],
) -> None:
    """Copy common files to the destination directory."""
    lock_files = set(LOCK_FILE_BY_MANAGER.values())
    current_lock = LOCK_FILE_BY_MANAGER.get(package_manager)
    for source in COMMON_DIR.iterdir():
        if source.is_dir() or source.name == ".DS_Store":
            continue

        name = source.name
        is_template = source.suffix == ".jinja2"

        if is_template:
            name = source.stem

        if orm_type == "tortoise" and name == "alembic.ini":
            continue
        if name == current_lock:
            continue
        if name in lock_files:
            continue

        target = destination / name
        if is_template:
            _render_template(source, target, context)
        else:
            target.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(source, target)


def _copy_src_app(destination: Path, orm_type: str, design: str) -> None:
    """Copy the application source directory to the destination."""
    target_dir = destination / "src" / "app"
    target_dir.parent.mkdir(parents=True, exist_ok=True)

    source_app_dir = SRC_DIR / design
    infra_dir = (source_app_dir / "infrastructure").resolve()
    models_dir = (source_app_dir / "models").resolve()

    def copy_tree_with_skip(
        source: Path, target: Path, skip_map: Mapping[Path, set[str]]
    ) -> None:
        def ignore(current: str, names: Iterable[str]) -> Iterable[str]:
            current_path = Path(current).resolve()
            for root, skip_names in skip_map.items():
                if current_path == root:
                    return [name for name in names if name in skip_names]
            return []

        shutil.copytree(source, target, dirs_exist_ok=True, ignore=ignore)

    if design == "ddd":
        skip = {infra_dir: set(ORM_CHOICES)}
        copy_tree_with_skip(source_app_dir, target_dir, skip)

        source_database = infra_dir / orm_type
        target_database = target_dir / "infrastructure" / "database"
        shutil.copytree(source_database, target_database, dirs_exist_ok=True)

    elif design == "mvc":
        skip = {source_app_dir.resolve(): {"models"}}
        copy_tree_with_skip(source_app_dir, target_dir, skip)

        source_models = models_dir / orm_type
        target_models = target_dir / "models"
        shutil.copytree(source_models, target_models, dirs_exist_ok=True)


def _resolve_compose_file(base: str, extension: str, orm_type: str) -> Path:
    """Resolve the compose file path for the given ORM type."""
    candidates = (
        COMPOSE_APP_DIR / f"{base}.{orm_type}.{extension}",
        COMPOSE_APP_DIR / f"{base}_{orm_type}.{extension}",
    )
    for candidate in candidates:
        if candidate.exists():
            return candidate
    raise FileNotFoundError(
        f"Could not find compose script for '{base}' with ORM '{orm_type}'."
    )


def _copy_compose_app(
    destination: Path, orm_type: str, context: Mapping[str, str]
) -> None:
    """Copy the compose app files to the destination directory."""
    target_dir = destination / "compose" / "app"

    def ignore(source: str, names: Iterable[str]) -> Iterable[str]:
        source_path = Path(source).resolve()
        if source_path == COMPOSE_APP_DIR:
            templates = {
                "dev.sqlalchemy.sh",
                "dev.tortoise.sh",
                "prod.sqlalchemy.py",
                "prod.tortoise.py",
                "Dockerfile.jinja2",
            }
            return [name for name in names if name in templates]
        return []

    shutil.copytree(
        COMPOSE_APP_DIR, target_dir, dirs_exist_ok=True, ignore=ignore
    )

    dev_source = _resolve_compose_file("dev", "sh", orm_type)
    prod_source = _resolve_compose_file("prod", "py", orm_type)
    shutil.copy2(dev_source, target_dir / "dev.sh")
    shutil.copy2(prod_source, target_dir / "prod.py")

    dockerfile_source = COMPOSE_APP_DIR / "Dockerfile.jinja2"
    _render_template(dockerfile_source, target_dir / "Dockerfile", context)


def copy_template(
    destination: Path,
    orm_type: str,
    design: str,
    project_name: str,
    package_manager: str,
) -> None:
    """Copy the complete template to the destination directory."""
    context = _get_template_config(
        design, orm_type, project_name, package_manager
    )
    _copy_src_app(destination, orm_type, design)
    _copy_compose_app(destination, orm_type, context)
    _copy_common_files(destination, orm_type, package_manager, context)


def collect_existing_items(destination: Path) -> set[Path]:
    """Return the set of existing items in a destination for cleanup logic."""
    return _collect_existing_items(destination)


def get_generated_items(
    orm_type: str, design: str, package_manager: str
) -> set[Path]:
    """Return the set of items this template would generate."""
    return _collect_generated_items(orm_type, design, package_manager)


def ensure_package_manager_available(package_manager: str) -> None:
    """Validate the requested package manager is available on PATH."""
    if package_manager not in PACKAGE_MANAGER_CHOICES:
        raise click.ClickException(
            "Unsupported package manager "
            f"'{package_manager}'. Choose from: {', '.join(PACKAGE_MANAGER_CHOICES)}"
        )

    if shutil.which(package_manager) is None:
        download_url = PACKAGE_MANAGER_DOWNLOAD_URLS.get(package_manager)
        download_hint = (
            f" Download it from {download_url} before running the create command."
            if download_url
            else " Install it before running the create command."
        )
        raise click.ClickException(
            f"Package manager '{package_manager}' is not installed.{download_hint}"
        )


def apply_package_manager(destination: Path, package_manager: str) -> None:
    """Generate the lock file for the project using the chosen package manager."""
    ensure_package_manager_available(package_manager)

    lock_file = LOCK_FILE_BY_MANAGER.get(package_manager, "lock file")
    lock_path = destination / lock_file
    env = os.environ.copy()
    if package_manager == "uv":
        env.setdefault("UV_CACHE_DIR", str(destination / ".uv-cache"))
    else:
        env.setdefault("POETRY_CACHE_DIR", str(destination / ".poetry-cache"))
    if package_manager == "uv":
        command = ["uv", "lock"]
    else:
        command = ["poetry", "lock", "--no-interaction", "--quiet"]

    result = subprocess.run(
        command, cwd=destination, capture_output=True, text=True, env=env
    )
    if result.returncode != 0:
        message = (
            result.stderr.strip() or result.stdout.strip() or "Unknown error"
        )
        raise click.ClickException(
            f"Failed to generate {lock_file} with {package_manager}: {message}"
        )
    if lock_file and not lock_path.exists():
        output = result.stdout.strip() or result.stderr.strip()
        detail = f" ({output})" if output else ""
        raise click.ClickException(
            f"{package_manager} finished without creating {lock_file}{detail}"
        )



from . import services  # noqa: F401
from .entities import (  # noqa: F401
    EmailChange,
    PasswordForgot,
    UserFlat,
    UserUncommitted,
)
from .repository import UsersRepository  # noqa: F401



from pydantic import BaseModel


class DocsSettings(BaseModel):
    docs: str = "/docs"
    redoc: str = "/redoc"


class Settings(BaseModel):
    domain: str = "http://backend.local"
    name: str = "Robyn Quick"
    version: str = "0.1.0"
    urls: DocsSettings = DocsSettings()



from pydantic import BaseModel


class Settings(BaseModel):
    host: str = "cache"
    port: int = 6379
    db: int = 0
    use_fake: bool = True
    ttl_activation_seconds: int = 3600
    ttl_password_reset_seconds: int = 3600



"""Runtime entrypoint for production images.

Runs Aerich migrations before handing off to the Robyn server (or any custom command).
"""

from __future__ import annotations

import os
import shlex
import subprocess
import sys

IGNORABLE_WARNINGS = ("App 'models' is already initialized.",)
APP_MODULE = os.environ.get("ROBYN_APP_MODULE", "app.server")


def _run(cmd: list[str], *, ignore_existing: bool = False) -> None:
    if not ignore_existing:
        subprocess.run(cmd, check=True)
        return

    proc = subprocess.run(cmd, capture_output=True, text=True)
    if proc.returncode == 0:
        sys.stdout.write(proc.stdout)
        sys.stderr.write(proc.stderr)
        return

    combined = f"{proc.stdout}{proc.stderr}"
    if any(warning in combined for warning in IGNORABLE_WARNINGS):
        sys.stdout.write(proc.stdout)
        sys.stderr.write(proc.stderr)
        return

    raise subprocess.CalledProcessError(
        proc.returncode, cmd, output=proc.stdout, stderr=proc.stderr
    )


def main() -> None:
    _run(["aerich", "init-db"], ignore_existing=True)
    _run(["aerich", "upgrade"])

    raw_cmd = os.environ.get("APP_CMD")
    if raw_cmd:
        _run(shlex.split(raw_cmd))
        return

    _run(
        [
            sys.executable,
            "-m",
            APP_MODULE,
            "--fast",
            "--processes",
            "4",
            "--workers",
            "3",
            "--log-level",
            "WARNING",
        ]
    )


if __name__ == "__main__":
    main()



"""Application settings inspired by src 2 design."""

from pathlib import Path

from pydantic_settings import BaseSettings, SettingsConfigDict

from . import authentication as _authentication
from . import cache as _cache
from . import core
from . import cors as _cors
from . import database as _database
from . import integrations as _integrations
from . import logging as _logging
from . import mailing as _mailing
from . import public_api as _public_api

__all__ = ("settings", "Settings")


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="SETTINGS__",
        env_nested_delimiter="__",
        env_file=".env",
        extra="ignore",
    )

    # infrastructure
    database: _database.Settings = _database.Settings()
    cache: _cache.Settings = _cache.Settings()

    # platform
    root_dir: Path
    src_dir: Path
    debug: bool = True
    public_api: _public_api.Settings = _public_api.Settings()
    logging: _logging.Settings = _logging.Settings()
    mailing: _mailing.Settings = _mailing.Settings()
    authentication: _authentication.Settings = _authentication.Settings()
    cors: _cors.Settings = _cors.Settings()

    # integrations
    integrations: _integrations.Settings = _integrations.Settings()


settings = Settings(root_dir=core.ROOT_PATH, src_dir=core.SRC_PATH)



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import AsyncGenerator

from loguru import logger
from tortoise.backends.base.client import BaseDBAsyncClient
from tortoise.exceptions import IntegrityError, OperationalError
from tortoise.transactions import in_transaction

from ....infrastructure.application import DatabaseError
from .engine import create_engine
from .session import CTX_CONNECTION


@asynccontextmanager
async def transaction() -> AsyncGenerator[BaseDBAsyncClient, None]:
    await create_engine()
    async with in_transaction(connection_name="default") as connection:
        token = CTX_CONNECTION.set(connection)
        try:
            yield connection
        except DatabaseError as exc:
            logger.error(f"Rolling back changes: {exc}")
            raise
        except (IntegrityError, OperationalError) as exc:  # pragma: no cover
            logger.error(f"Rolling back changes: {exc}")
            raise DatabaseError(message=str(exc)) from exc
        finally:
            CTX_CONNECTION.reset(token)



"""Simple signed-cookie session middleware for Robyn."""

from __future__ import annotations

import hashlib
import hmac
import json
from base64 import urlsafe_b64decode, urlsafe_b64encode
from contextvars import ContextVar
from dataclasses import dataclass
from typing import Any, Mapping, MutableMapping, cast

from robyn import Request, Response, Robyn

from ....config import settings

SESSION_COOKIE_NAME = "robyn_session"
SESSION_MAX_AGE = 60 * 60 * 24 * 14  # 14 days


@dataclass
class SessionState:
    data: dict[str, Any]
    serialized: str


_state: ContextVar[SessionState | None] = ContextVar(
    "session_state", default=None
)


def get_session() -> dict[str, Any]:
    state = _state.get()
    return state.data if state else {}


def _extract_cookie(raw_cookie_header: str | None) -> str | None:
    if not raw_cookie_header:
        return None

    for chunk in raw_cookie_header.split(";"):
        name, _, value = chunk.strip().partition("=")
        if name == SESSION_COOKIE_NAME:
            return value or None
    return None


def _encode_session(session: dict[str, Any], secret: bytes) -> str:
    payload = json.dumps(
        session, separators=(",", ":"), sort_keys=True
    ).encode()
    signature = hmac.new(secret, payload, hashlib.sha256).hexdigest()
    return f"{urlsafe_b64encode(payload).decode()}.{signature}"


def _decode_session(value: str, secret: bytes) -> tuple[dict[str, Any], str]:
    try:
        payload_b64, signature = value.split(".", 1)
        payload = urlsafe_b64decode(_pad_base64(payload_b64))
        expected = hmac.new(secret, payload, hashlib.sha256).hexdigest()
        if not hmac.compare_digest(signature, expected):
            return {}, ""
        data = json.loads(payload)
        if isinstance(data, dict):
            return data, value
    except Exception:
        pass
    return {}, ""


def _pad_base64(value: str) -> bytes:
    padding = (-len(value)) % 4
    return (value + ("=" * padding)).encode()


def _cookie_header(serialized: str | None) -> str:
    base = f"{SESSION_COOKIE_NAME}={serialized or ''}"
    parts = [base, "Path=/", "HttpOnly", "SameSite=Lax"]
    if serialized is None:
        parts.append("Max-Age=0")
    else:
        parts.append(f"Max-Age={SESSION_MAX_AGE}")
    return "; ".join(parts)


def register(app: Robyn) -> None:
    secret = settings.authentication.session_secret_key.encode()

    @app.before_request()
    def load_session(request: Request):
        raw_headers = getattr(request, "headers", {}) or {}
        if isinstance(raw_headers, dict):
            header_map = raw_headers
        else:  # Robyn exposes Headers objects; convert defensively
            try:
                header_map = dict(raw_headers)  # type: ignore[arg-type]
            except Exception:  # pragma: no cover - safety net
                header_map = {}

        raw_cookie = None
        for key, value in header_map.items():
            if isinstance(key, bytes):
                key = key.decode()
            if key.lower() == "cookie":
                raw_cookie = value if isinstance(value, str) else str(value)
                break

        cookie_value = _extract_cookie(raw_cookie)
        session_data: dict[str, Any]
        serialized: str
        if cookie_value:
            session_data, serialized = _decode_session(cookie_value, secret)
        else:
            session_data, serialized = {}, ""

        state = SessionState(data=session_data, serialized=serialized)
        _state.set(state)
        # Robyn Request objects don't allow setting arbitrary attributes, so
        # consumers should import `get_session()` from this module instead.
        return request

    @app.after_request()
    def persist_session(response: Response):
        state = _state.get()
        if state is None:
            return response

        headers_source = response.headers
        if isinstance(headers_source, MutableMapping):
            headers = dict(headers_source.items())
        elif headers_source is None:
            headers = {}
        else:
            headers = dict(cast(Mapping[str, Any], headers_source).items())
        session_data = state.data
        serialized = state.serialized

        if not session_data:
            if serialized:
                headers["set-cookie"] = _cookie_header(None)
        else:
            new_serialized = _encode_session(session_data, secret)
            if new_serialized != serialized:
                headers["set-cookie"] = _cookie_header(new_serialized)

        if headers:
            response.headers = headers

        _state.set(None)
        return response



from pydantic import BaseModel


class Settings(BaseModel):
    format: str = "{time:YYYY-MM-DD HH:mm:ss} | {level:<5} | {message}"
    file: str = "robyn_app"
    rotation: str = "10 MB"
    compression: str = "zip"



from __future__ import annotations

from email.message import EmailMessage as SMTPEmailMessage
from email.utils import formataddr
from typing import Sequence

import aiosmtplib
from loguru import logger
from pydantic import EmailStr

from .config import settings
from .schemas import InternalEntity


class EmailMessage(InternalEntity):
    recipients: Sequence[EmailStr]
    subject: str
    body: str


class MailingService:
    def __init__(self) -> None:
        self._config = settings.mailing

    def _build_message(self, message: EmailMessage) -> SMTPEmailMessage:
        email = SMTPEmailMessage()
        sender = (
            formataddr((self._config.sender_name, self._config.sender_email))
            if self._config.sender_name
            else str(self._config.sender_email)
        )
        email["From"] = sender
        email["To"] = ", ".join(message.recipients)
        email["Subject"] = message.subject
        email.set_content(message.body)
        return email

    async def send(self, message: EmailMessage) -> None:
        smtp_message = self._build_message(message)
        logger.info(
            "Sending email via SMTP -> host=%s port=%s recipients=%s subject=%s",
            self._config.host,
            self._config.port,
            ",".join(message.recipients),
            message.subject,
        )
        await aiosmtplib.send(
            smtp_message,
            hostname=self._config.host,
            port=self._config.port,
            start_tls=self._config.start_tls,
            username=self._config.username,
            password=self._config.password,
        )


mailing_service = MailingService()



from datetime import datetime

from pydantic import Field

from .base import InternalEntity


class TimeStampMixin(InternalEntity):
    created_at: datetime = Field(default_factory=datetime.utcnow)
    updated_at: datetime = Field(default_factory=datetime.utcnow)



"""Factory helpers for creating configured Robyn applications."""

from collections.abc import Iterable
from typing import Callable

from robyn import Response, Robyn
from robyn.authentication import AuthenticationHandler

RouteRegistrar = Callable[[Robyn], None]
MiddlewareRegistrar = Callable[[Robyn], None]


def create(
    import_name: str,
    *,
    route_registrars: Iterable[RouteRegistrar] | None = None,
    middlewares: Iterable[MiddlewareRegistrar] | None = None,
    exception_handler: Callable[[Exception], Response] | None = None,
    authentication_handler: AuthenticationHandler | None = None,
    **robyn_kwargs,
) -> Robyn:
    """Create a Robyn application with optional route and middleware registrars."""

    app = Robyn(import_name, **robyn_kwargs)

    if exception_handler is not None:
        app.exception(exception_handler)

    if authentication_handler is not None:
        app.configure_authentication(authentication_handler)

    if middlewares is not None:
        for middleware in middlewares:
            middleware(app)

    if route_registrars is not None:
        for register in route_registrars:
            register(app)

    return app



"""Application services orchestrating domain + infrastructure, Robyn edition."""

import uuid
from typing import Any

from pydantic import EmailStr

from ..config import settings
from ..domain.users import EmailChange, UserFlat
from ..domain.users import services as users_services
from ..infrastructure.application import NotFoundError, UnprocessableError
from ..infrastructure.authentication import AuthProvider, pwd_context
from ..infrastructure.cache import CacheRepository
from ..infrastructure.database import transaction
from ..infrastructure.database.repository import (
    UsersRepository as InfrastructureUsersRepository,
)
from ..infrastructure.mailing import EmailMessage, mailing_service

__all__ = (
    "create",
    "create_external",
    "activate",
    "password_change",
    "password_reset",
    "password_reset_change",
    "email_change_request",
    "email_change_confirmation",
    "get_by_login",
)


async def create(payload: dict[str, Any]) -> UserFlat:
    plain_password = payload.pop("password")
    payload["password"] = pwd_context.hash(plain_password, scheme="bcrypt")
    user = await users_services.create(
        payload=payload,
        repository_factory=InfrastructureUsersRepository,
    )

    activation_key = users_services.generate_key(user.email)
    activation_link = users_services.create_activation_link(activation_key)

    async with CacheRepository[UserFlat]() as cache:
        await cache.set(
            namespace="activation",
            key=activation_key,
            instance=user,
            ttl=settings.cache.ttl_activation_seconds or None,
        )

    await mailing_service.send(
        EmailMessage(
            recipients=[user.email],
            subject="Account activation",
            body=f"Activation link: {activation_link}",
        )
    )

    return user


async def create_external(payload: dict[str, Any]) -> UserFlat:
    return await users_services.create(
        payload=payload,
        repository_factory=InfrastructureUsersRepository,
    )


async def update_partial(user: UserFlat, payload: dict[str, Any]) -> UserFlat:
    return await users_services.update_partial(
        id_=user.id,
        payload=payload,
        repository_factory=InfrastructureUsersRepository,
    )


async def get_by_login(login: str) -> UserFlat:
    async with transaction():
        return await InfrastructureUsersRepository().get_by_login(login=login)


async def get(user_id: int) -> UserFlat:
    async with transaction():
        return await InfrastructureUsersRepository().get(id_=user_id)


async def activate(key: uuid.UUID) -> UserFlat:
    async with CacheRepository[UserFlat]() as cache:
        cache_entry = await cache.get(namespace="activation", key=key)
        user = await users_services.activate(
            cache_entry.instance.id,
            repository_factory=InfrastructureUsersRepository,
        )
        await cache.delete(namespace="activation", key=key)
        return user


async def password_change(
    user: UserFlat, old_password: str, new_password: str
) -> UserFlat:
    if not pwd_context.verify(old_password, user.password, scheme="bcrypt"):
        raise UnprocessableError(message="Password invalid")

    password_hash = pwd_context.hash(new_password, scheme="bcrypt")
    return await users_services.password_update(
        id_=user.id,
        password_hash=password_hash,
        repository_factory=InfrastructureUsersRepository,
    )


async def password_reset(email: EmailStr) -> None:
    try:
        async with transaction():
            repository = InfrastructureUsersRepository()
            user = await repository.get_by_login(login=email)
    except NotFoundError:
        return

    reset_key = users_services.generate_key(email)
    reset_link = users_services.create_password_reset_link(reset_key)

    async with CacheRepository[UserFlat]() as cache:
        await cache.set(
            namespace="password-reset",
            key=reset_key,
            instance=user,
            ttl=settings.cache.ttl_password_reset_seconds or None,
        )

    await mailing_service.send(
        EmailMessage(
            recipients=[email],
            subject="Password reset",
            body=f"Reset link: {reset_link}",
        )
    )


async def password_reset_change(key: uuid.UUID, new_password: str) -> UserFlat:
    async with CacheRepository[UserFlat]() as cache:
        try:
            cache_entry = await cache.get(namespace="password-reset", key=key)
        except NotFoundError as exc:
            raise UnprocessableError(message="Invalid key") from exc

        password_hash = pwd_context.hash(new_password, scheme="bcrypt")
        user = await users_services.password_update(
            id_=cache_entry.instance.id,
            password_hash=password_hash,
            repository_factory=InfrastructureUsersRepository,
        )
        await cache.delete(namespace="password-reset", key=key)
        return user


async def email_change_request(user: UserFlat, email: EmailStr) -> None:
    change_key = users_services.generate_key(email)
    change_link = users_services.create_email_change_link(change_key)

    async with CacheRepository[EmailChange]() as cache:
        await cache.set(
            namespace="email-change",
            key=change_key,
            instance=EmailChange(user_id=user.id, email=email),
        )

    await mailing_service.send(
        EmailMessage(
            recipients=[email],
            subject="Email migration",
            body=f"Confirmation link: {change_link}",
        )
    )


async def email_change_confirmation(key: uuid.UUID) -> UserFlat:
    async with CacheRepository[EmailChange]() as cache:
        try:
            cache_entry = await cache.get(namespace="email-change", key=key)
        except NotFoundError as exc:
            raise UnprocessableError(message="Invalid key") from exc

        async with transaction():
            repository = InfrastructureUsersRepository()
            user = await repository.update(
                attr="id",
                value=cache_entry.instance.user_id,
                payload={
                    "email": cache_entry.instance.email,
                    "auth_provider": AuthProvider.INTERNAL,
                },
            )
        await cache.delete(namespace="email-change", key=key)
        return user



from robyn import Robyn

from . import authentication, healthcheck, users


def register_routes(app: Robyn) -> None:
    healthcheck.register(app)
    users.register(app)
    authentication.register(app)



from __future__ import annotations

import asyncio
import functools
from typing import Any

from tortoise import Tortoise

from ....config import settings

APP_LABEL = "models"
MODEL_MODULES = (
    "app.infrastructure.database.tables",
    "aerich.models",
)
DRIVER_ALIASES = {
    "sqlite+aiosqlite": "sqlite",
    "postgresql+asyncpg": "postgres",
    "mysql+aiomysql": "mysql",
}

_ENGINE_LOCK = asyncio.Lock()
_INITIALIZED = False


def _normalize_database_url(url: str) -> str:
    for candidate, alias in DRIVER_ALIASES.items():
        if url.startswith(candidate):
            return url.replace(candidate, alias, 1)
    return url


def _build_config() -> dict[str, Any]:
    return {
        "connections": {
            "default": _normalize_database_url(settings.database.url)
        },
        "apps": {
            APP_LABEL: {
                "models": list(MODEL_MODULES),
                "default_connection": "default",
            }
        },
        "use_tz": False,
        "timezone": "UTC",
    }


@functools.lru_cache(maxsize=1)
def build_engine() -> dict[str, Any]:
    return _build_config()


TORTOISE_ORM = build_engine()


async def create_engine() -> dict[str, Any]:
    global _INITIALIZED
    if _INITIALIZED:
        return TORTOISE_ORM

    async with _ENGINE_LOCK:
        if not _INITIALIZED:
            await Tortoise.init(config=TORTOISE_ORM)
            _INITIALIZED = True
    return TORTOISE_ORM


async def close_engine() -> None:
    global _INITIALIZED
    if not _INITIALIZED:
        return
    await Tortoise.close_connections()
    _INITIALIZED = False



import uuid

from app.domain.authentication import validators as auth_validators
from app.infrastructure.application import PublicEntity, UnprocessableError
from pydantic import EmailStr, Field, field_validator, model_validator


class _BaseUser(PublicEntity):
    username: str = Field(
        description="User username",
        examples=["john"],
        min_length=1,
        max_length=255,
    )
    email: EmailStr = Field(
        description="User email",
        examples=["john@email.com"],
    )


class UserCreateBody(_BaseUser):
    password: str = Field(
        description="User password",
        examples=["@Dm1n#LKJ"],
    )

    @field_validator("password", mode="before")
    @classmethod
    def password_nist(cls, value: str) -> str:
        return auth_validators.password_nist(value)

    @model_validator(mode="before")
    @classmethod
    def username_email(cls, values: dict) -> dict:
        """Check if the username and email are not the same."""

        if values.get("username") == values.get("email"):
            raise UnprocessableError(
                message="The username and email must be different."
            )

        return values


class UserExternalBody(_BaseUser):
    role: int = Field(
        description="User role",
        examples=[1],
        default=1,
    )
    auth_provider: str | None = Field(
        default=None,
        description="External auth provider identifier",
    )


class ActivationBody(PublicEntity):
    key: uuid.UUID = Field(description="Activation key from email")


class PasswordChangeBody(PublicEntity):
    old_password: str = Field(description="User's current password")
    new_password: str = Field(description="A new user's password")

    @field_validator("new_password", mode="before")
    @classmethod
    def password_nist(cls, value: str) -> str:
        return auth_validators.password_nist(value)


class PasswordResetRequestBody(PublicEntity):
    email: EmailStr = Field(
        description="User email", examples=["john@email.com"]
    )


class PasswordResetConfirmBody(PublicEntity):
    key: uuid.UUID = Field(
        description="Password reset key that is taken from the email"
    )
    password: str = Field(description="A new user's password")

    @field_validator("password", mode="before")
    @classmethod
    def password_nist(cls, value: str) -> str:
        return auth_validators.password_nist(value)


class EmailChangeRequestBody(PublicEntity):
    email: EmailStr = Field(description="A new email you want to change to")


class EmailChangeConfirmBody(PublicEntity):
    key: uuid.UUID = Field(
        description="Email change key that is taken from the email"
    )


class UserPublic(_BaseUser):
    """The public user's data model."""

    id: int = Field(description="User id", examples=[1])
    is_active: bool = Field(
        description="Whether the user activated the account",
        examples=[True],
    )
    role: int = Field(
        description="User role. Possible values: [1,2,3]", examples=[2]
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, AsyncGenerator

from .entities import UserFlat, UserUncommitted


class UsersRepository(ABC):
    """Contract that captures user persistence operations."""

    @abstractmethod
    async def all(self) -> AsyncGenerator[UserFlat, None]:
        """Return every user instance."""

    @abstractmethod
    async def get(self, id_: int) -> UserFlat:
        """Return a specific user by identifier."""

    @abstractmethod
    async def get_by_login(self, login: str) -> UserFlat:
        """Return a user matching a login credential."""

    @abstractmethod
    async def create(self, schema: UserUncommitted) -> UserFlat:
        """Persist a new user from the provided schema."""

    @abstractmethod
    async def update(
        self, attr: str, value: Any, payload: dict[str, Any]
    ) -> UserFlat:
        """Update an existing user."""



"""Utility functions for the 'add' command."""

from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path
from typing import Any

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from jinja2 import Environment, StrictUndefined

ADD_MODULE_ROOT = Path(__file__).resolve().parent
JINJA_ENV = Environment(undefined=StrictUndefined)

# Default routes for code injection (overridable via [tool.robyn-config.add])
DEFAULT_DDD_DOMAIN_PATH = Path("src/app/domain")
DEFAULT_DDD_OPERATIONAL_PATH = Path("src/app/operational")
DEFAULT_DDD_PRESENTATION_PATH = Path("src/app/presentation")
DEFAULT_DDD_DB_REPOSITORY_PATH = Path(
    "src/app/infrastructure/database/repository"
)
DEFAULT_DDD_DB_TABLE_PATH = Path("src/app/infrastructure/database/tables.py")

DEFAULT_MVC_VIEWS_PATH = Path("src/app/views")
DEFAULT_MVC_DB_REPOSITORY_PATH = Path("src/app/models/repository.py")
DEFAULT_MVC_DB_TABLE_PATH = Path("src/app/models/models.py")
DEFAULT_MVC_URLS_PATH = Path("src/app/urls.py")


@dataclass
class DDDAddPaths:
    domain: Path
    operational: Path
    presentation: Path
    db_repository: Path
    db_tables: Path


@dataclass
class MVCAddPaths:
    views: Path
    db_repository: Path
    db_tables: Path
    urls: Path


def _normalize_entity_name(name: str) -> tuple[str, str]:
    """Normalize entity name to snake_case and PascalCase variants."""
    normalized = name.lower().replace("-", "_").replace(" ", "_")
    capitalized = "".join(word.capitalize() for word in normalized.split("_"))
    return normalized, capitalized


def _format_comment(comment: str) -> str:
    """Normalize inline comment formatting (adds leading space and #)."""
    if not comment:
        return ""

    cleaned = comment.strip()
    if not cleaned.startswith("#"):
        cleaned = f"# {cleaned}"
    return f" {cleaned}"


def read_project_config(project_path: Path) -> dict[str, Any]:
    """Read pyproject.toml and extract [tool.robyn-config] section."""
    pyproject_path = project_path / "pyproject.toml"

    if not pyproject_path.exists():
        raise FileNotFoundError(
            f"pyproject.toml not found in {project_path}. "
            "Make sure you're in a robyn-config project directory."
        )

    with open(pyproject_path, "rb") as f:
        data = tomllib.load(f)

    robyn_config = data.get("tool", {}).get("robyn-config", {})
    if not robyn_config:
        raise ValueError(
            "No [tool.robyn-config] section found in pyproject.toml. "
            "This project was not created with robyn-config."
        )

    return robyn_config


def _extract_design_orm(config: dict[str, Any]) -> tuple[str, str]:
    """Validate project has robyn-config metadata and return (design, orm)."""
    design = config.get("design")
    orm = config.get("orm")

    if not design or not orm:
        raise ValueError(
            "Invalid [tool.robyn-config] section. "
            "Missing 'design' or 'orm' fields."
        )
    return design, orm


def validate_project(project_path: Path) -> tuple[str, str]:
    """Validate project has robyn-config metadata and return (design, orm)."""
    config = read_project_config(project_path)
    return _extract_design_orm(config)


def _resolve_path(
    project_root: Path, raw_value: str | None, default: Path
) -> Path:
    """Resolve a configured path (relative to project root) or fall back to default."""
    return (
        (project_root / raw_value) if raw_value else (project_root / default)
    )


def _load_add_paths(
    project_path: Path, design: str, config: dict[str, Any]
) -> DDDAddPaths | MVCAddPaths:
    """Resolve add-paths from pyproject config (with defaults)."""
    add_config = config.get("add") or {}
    if design == "ddd":
        return DDDAddPaths(
            domain=_resolve_path(
                project_path,
                add_config.get("domain_path"),
                DEFAULT_DDD_DOMAIN_PATH,
            ),
            operational=_resolve_path(
                project_path,
                add_config.get("operational_path"),
                DEFAULT_DDD_OPERATIONAL_PATH,
            ),
            presentation=_resolve_path(
                project_path,
                add_config.get("presentation_path"),
                DEFAULT_DDD_PRESENTATION_PATH,
            ),
            db_repository=_resolve_path(
                project_path,
                add_config.get("database_repository_path"),
                DEFAULT_DDD_DB_REPOSITORY_PATH,
            ),
            db_tables=_resolve_path(
                project_path,
                add_config.get("database_table_path"),
                DEFAULT_DDD_DB_TABLE_PATH,
            ),
        )

    if design == "mvc":
        return MVCAddPaths(
            views=_resolve_path(
                project_path,
                add_config.get("views_path"),
                DEFAULT_MVC_VIEWS_PATH,
            ),
            db_repository=_resolve_path(
                project_path,
                add_config.get("database_repository_path"),
                DEFAULT_MVC_DB_REPOSITORY_PATH,
            ),
            db_tables=_resolve_path(
                project_path,
                add_config.get("database_table_path"),
                DEFAULT_MVC_DB_TABLE_PATH,
            ),
            urls=_resolve_path(
                project_path,
                add_config.get("urls_path"),
                DEFAULT_MVC_URLS_PATH,
            ),
        )

    raise ValueError(f"Unsupported design pattern: {design}")


def _render_template_file(
    source: Path, target: Path, context: dict[str, str]
) -> None:
    """Render a Jinja2 template file to target location."""
    template_content = source.read_text()
    template = JINJA_ENV.from_string(template_content)
    rendered = template.render(**context)
    target.parent.mkdir(parents=True, exist_ok=True)
    target.write_text(rendered)


def _render_template_string(template_str: str, context: dict[str, str]) -> str:
    """Render a Jinja2 template string."""
    template = JINJA_ENV.from_string(template_str)
    return template.render(**context)


def _render_templates_from_directory(
    template_dir: Path,
    target_dir: Path,
    context: dict[str, str],
    project_root: Path,
    created_files: list[str],
) -> None:
    """Render all Jinja2 templates in a directory to the target directory."""
    for template_file in template_dir.glob("*.jinja2"):
        target_path = target_dir / template_file.stem
        _render_template_file(template_file, target_path, context)
        created_files.append(str(target_path.relative_to(project_root)))


def _update_init_file(
    init_path: Path, import_line: str, export_name: str
) -> None:
    """Add import (and optionally __all__ export) to a module file."""
    init_path.parent.mkdir(parents=True, exist_ok=True)
    content = init_path.read_text() if init_path.exists() else ""

    if import_line not in content:
        lines = content.split("\n") if content else []
        insert_pos = 0
        for i, line in enumerate(lines):
            if line.startswith("from ") or line.startswith("import "):
                insert_pos = i + 1
        lines.insert(insert_pos, import_line)
        content = "\n".join(lines)
        init_path.write_text(content)

    if export_name:
        content = init_path.read_text()
        if "__all__" in content:
            _add_to_all_list(init_path, export_name)
        else:
            suffix = f'\n__all__ = (\n    "{export_name}",\n)\n'
            if content and not content.endswith("\n"):
                suffix = "\n" + suffix.lstrip("\n")
            init_path.write_text(f"{content}{suffix}")


def _find_closing_parenthesis(lines: list[str], start_index: int) -> int:
    """Locate the index of the closing parenthesis in a multiline import."""
    for idx in range(start_index, len(lines)):
        if ")" in lines[idx]:
            return idx
    return len(lines)


def _detect_indent(lines: list[str], closing_index: int) -> str:
    """Detect indentation level for items inside a multiline import."""
    if closing_index >= len(lines):
        return "    "

    closing_line = lines[closing_index]
    match = re.match(r"(\s*)\)", closing_line)
    if match:
        return match.group(1) or "    "

    match = re.match(r"(\s*)", closing_line)
    return (match.group(1) if match else "") or "    "


def _append_inline_paren_import(
    line: str, import_item: str, trailing_comment: str
) -> str:
    """Append an import item to a single-line parenthesized import."""
    comment = ""
    base_line = line
    if "#" in line:
        base_line, existing_comment = line.split("#", 1)
        comment = "  #" + existing_comment

    before_paren, after_paren = base_line.split("(", 1)
    inside, after = after_paren.split(")", 1)
    inside = inside.strip()
    updated_inside = f"{inside}, {import_item}" if inside else f"{import_item}"
    comment = comment or _format_comment(trailing_comment)
    return f"{before_paren}({updated_inside}){after}{comment}"


def _ensure_import_from(
    file_path: Path,
    module: str,
    import_item: str,
    *,
    trailing_comment: str = "",
) -> None:
    """Ensure `from {module} import {import_item}` exists in file_path."""
    formatted_comment = _format_comment(trailing_comment)

    if not file_path.exists():
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.write_text(
            f"from {module} import {import_item}{formatted_comment}\n"
        )
        return

    lines = file_path.read_text().split("\n")
    prefix = f"from {module} import "

    for i, line in enumerate(lines):
        stripped = line.strip()
        if not stripped.startswith(prefix):
            continue

        if "(" in stripped and ")" in stripped:
            inside = stripped.split("(", 1)[1].rsplit(")", 1)[0]
            if re.search(rf"\b{re.escape(import_item)}\b", inside):
                return
            lines[i] = _append_inline_paren_import(
                line, import_item, trailing_comment
            )
            file_path.write_text("\n".join(lines))
            return

        if "(" in stripped:
            closing_index = _find_closing_parenthesis(lines, i)
            block = "\n".join(lines[i : closing_index + 1])
            if re.search(rf"\b{re.escape(import_item)}\b", block):
                return
            indent = _detect_indent(lines, closing_index)
            lines.insert(closing_index, f"{indent}{import_item},")
            file_path.write_text("\n".join(lines))
            return

        if re.search(rf"\b{re.escape(import_item)}\b", stripped):
            return

        comment = ""
        if "#" in line:
            base, existing_comment = line.split("#", 1)
            line = base.rstrip()
            comment = "  #" + existing_comment
        elif formatted_comment:
            comment = formatted_comment

        lines[i] = f"{line.rstrip()}, {import_item}{comment}"
        file_path.write_text("\n".join(lines))
        return

    insert_pos = 0
    for i, line in enumerate(lines):
        if line.startswith("from ") or line.startswith("import "):
            insert_pos = i + 1
    lines.insert(insert_pos, f"{prefix}{import_item}{formatted_comment}")
    file_path.write_text("\n".join(lines))


def _ensure_register_call(target_file: Path, register_call: str) -> None:
    """Ensure register(app) call is present after existing registrations."""
    if not target_file.exists():
        return

    lines = target_file.read_text().split("\n")
    stripped_call = register_call.strip()
    if stripped_call in (line.strip() for line in lines):
        return

    inserted = False
    for i in range(len(lines) - 1, -1, -1):
        if lines[i].strip().endswith(".register(app)"):
            lines.insert(i + 1, register_call)
            inserted = True
            break

    if inserted:
        target_file.write_text("\n".join(lines))


def _add_table_to_tables_py(
    tables_path: Path,
    name: str,
    name_capitalized: str,
    orm: str,
    context: dict[str, str],
) -> None:
    """Add table class to tables.py file."""
    if not tables_path.exists():
        return

    content = tables_path.read_text()
    table_class_name = f"{name_capitalized}Table"

    # Check if table already exists
    if table_class_name in content:
        return

    # Get the table template
    template_file = (
        ADD_MODULE_ROOT
        / "ddd"
        / "infrastructure"
        / orm
        / "__name___table.py.jinja2"
    )
    if not template_file.exists():
        return

    template_content = template_file.read_text()
    # Extract just the class definition (skip imports)
    lines = template_content.split("\n")
    class_start = None
    for i, line in enumerate(lines):
        if line.startswith("class "):
            class_start = i
            break

    if class_start is not None:
        class_definition = "\n".join(lines[class_start:])
        rendered_class = _render_template_string(class_definition, context)

        # Add to end of file
        if not content.endswith("\n"):
            content += "\n"
        content += f"\n\n{rendered_class}\n"
        tables_path.write_text(content)

        # Update __all__ if it exists
        if "__all__" in content:
            # Use regex to add to __all__ tuple
            content = tables_path.read_text()
            all_pattern = r"(__all__\s*=\s*\(\s*)(.*?)(\s*\))"
            match = re.search(all_pattern, content, re.DOTALL)
            if match:
                current_items = match.group(2).strip()
                if current_items.endswith(","):
                    new_items = f'{current_items}\n    "{table_class_name}",'
                else:
                    new_items = f'{current_items},\n    "{table_class_name}",'
                new_all = f"{match.group(1)}{new_items}{match.group(3)}"
                content = (
                    content[: match.start()] + new_all + content[match.end() :]
                )
                tables_path.write_text(content)


def _register_routes_ddd(presentation_path: Path, name: str) -> None:
    """Register routes in DDD presentation/__init__.py."""
    pres_init = presentation_path / "__init__.py"
    if not pres_init.exists():
        return

    _ensure_import_from(pres_init, ".", name)
    _ensure_register_call(pres_init, f"    {name}.register(app)")


def _register_routes_mvc(urls_path: Path, name: str) -> None:
    """Register routes in MVC urls.py."""
    if not urls_path.exists():
        return

    _ensure_import_from(urls_path, ".views", name)
    _ensure_register_call(urls_path, f"    {name}.register(app)")


def _add_ddd_templates(
    project_path: Path,
    paths: DDDAddPaths,
    name: str,
    name_capitalized: str,
    orm: str,
) -> list[str]:
    """Add DDD templates to the project."""
    templates_path = ADD_MODULE_ROOT / "ddd"
    created_files = []

    context = {
        "name": name,
        "Name": name_capitalized,
        "orm": orm,
    }

    # Domain layer
    domain_dir = paths.domain / name
    domain_templates = templates_path / "domain" / "__name__"
    _render_templates_from_directory(
        domain_templates, domain_dir, context, project_path, created_files
    )

    # Update domain __init__.py
    domain_init = paths.domain / "__init__.py"
    _ensure_import_from(
        domain_init, ".", name, trailing_comment="# noqa: F401"
    )

    # Add table to tables.py
    _add_table_to_tables_py(
        paths.db_tables, name, name_capitalized, orm, context
    )

    # Infrastructure repository
    repo_template = (
        templates_path
        / "infrastructure"
        / orm
        / "repository"
        / "__name__.py.jinja2"
    )
    if repo_template.exists():
        repo_target = paths.db_repository / f"{name}.py"
        _render_template_file(repo_template, repo_target, context)
        created_files.append(str(repo_target.relative_to(project_path)))

        # Update repository __init__.py
        repo_init = paths.db_repository / "__init__.py"
        _update_init_file(
            repo_init,
            f"from .{name} import {name_capitalized}Repository  # noqa: F401",
            f"{name_capitalized}Repository",
        )

    # Operational layer
    ops_template = templates_path / "operational" / "__name__.py.jinja2"
    if ops_template.exists():
        ops_target = paths.operational / f"{name}.py"
        _render_template_file(ops_template, ops_target, context)
        created_files.append(str(ops_target.relative_to(project_path)))

        # Update operational __init__.py
        ops_init = paths.operational / "__init__.py"
        _ensure_import_from(
            ops_init, ".", name, trailing_comment="# noqa: F401"
        )

    # Presentation layer
    pres_dir = paths.presentation / name
    pres_templates = templates_path / "presentation" / "__name__"
    _render_templates_from_directory(
        pres_templates, pres_dir, context, project_path, created_files
    )

    # Auto-register routes in presentation/__init__.py
    _register_routes_ddd(paths.presentation, name)

    return created_files


def _append_class_to_file(
    file_path: Path,
    template_path: Path,
    context: dict[str, str],
    class_name: str,
) -> None:
    """Append a class definition from a template to a file if it doesn't exist."""
    if not file_path.exists() or not template_path.exists():
        return

    content = file_path.read_text()
    if class_name in content:
        return

    class_def = _render_template_string(template_path.read_text(), context)

    if not content.endswith("\n"):
        content += "\n"
    content += f"\n{class_def}\n"
    file_path.write_text(content)


def _add_to_all_list(file_path: Path, item_name: str) -> None:
    """Add an item to the __all__ tuple in a file."""
    if not file_path.exists():
        return

    content = file_path.read_text()
    all_pattern = r"(__all__\s*=\s*\(\s*)(.*?)(\s*\))"
    match = re.search(all_pattern, content, re.DOTALL)
    if match and item_name not in match.group(2):
        current_items = match.group(2).strip()
        if current_items and not current_items.endswith(","):
            new_items = f'{current_items},\n    "{item_name}",'
        else:
            new_items = f'{current_items}\n    "{item_name}",'
        new_all = f"{match.group(1)}{new_items}{match.group(3)}"
        content = content[: match.start()] + new_all + content[match.end() :]
        file_path.write_text(content)


def _add_mvc_templates(
    project_path: Path,
    paths: MVCAddPaths,
    name: str,
    name_capitalized: str,
    orm: str,
) -> list[str]:
    """Add MVC templates to the project."""
    templates_path = ADD_MODULE_ROOT / "mvc"
    created_files = []

    context = {
        "name": name,
        "Name": name_capitalized,
        "orm": orm,
    }

    # Models layer - Append to existing files
    models_file = paths.db_tables
    repo_file = paths.db_repository

    # 1. Append Table to models.py
    table_template = templates_path / "models" / orm / "table.py.jinja2"
    table_class = f"{name_capitalized}Table"
    if models_file.exists() and table_template.exists():
        _append_class_to_file(
            models_file, table_template, context, table_class
        )
        _add_to_all_list(models_file, table_class)
        created_files.append(str(models_file.relative_to(project_path)))

    # 2. Append Repository to repository.py
    repo_template = templates_path / "models" / orm / "repository.py.jinja2"
    repo_class = f"{name_capitalized}Repository"
    if repo_file.exists() and repo_template.exists():
        _ensure_import_from(repo_file, ".models", table_class)

        _append_class_to_file(repo_file, repo_template, context, repo_class)
        created_files.append(str(repo_file.relative_to(project_path)))

        # Update models __init__.py to export Repository
        models_init = paths.db_tables.parent / "__init__.py"
        _update_init_file(
            models_init,
            f"from .repository import {repo_class}  # noqa: F401",
            repo_class,
        )

    # Views layer
    views_template = templates_path / "views" / "__name__.py.jinja2"
    if views_template.exists():
        views_target = paths.views / f"{name}.py"
        _render_template_file(views_template, views_target, context)
        created_files.append(str(views_target.relative_to(project_path)))

        # Update views __init__.py
        views_init = paths.views / "__init__.py"
        _update_init_file(
            views_init,
            f"from .{name} import register as register_{name}  # noqa: F401",
            f"register_{name}",
        )

    # Auto-register routes in urls.py
    _register_routes_mvc(paths.urls, name)

    return created_files


def add_business_logic(project_path: Path, name: str) -> list[str]:
    """Add business logic templates to an existing project."""
    config = read_project_config(project_path)
    design, orm = _extract_design_orm(config)
    add_paths = _load_add_paths(project_path, design, config)
    name_lower, name_capitalized = _normalize_entity_name(name)

    if design == "ddd":
        return _add_ddd_templates(
            project_path, add_paths, name_lower, name_capitalized, orm
        )
    elif design == "mvc":
        return _add_mvc_templates(
            project_path, add_paths, name_lower, name_capitalized, orm
        )
    else:
        raise ValueError(f"Unsupported design pattern: {design}")
